"""ESI client for standingssync."""

from pathlib import Path

from esi.openapi_clients import ESIClientProvider

from standingssync import __version__

spec_file = Path(__file__).parent / "openapi_2025-12-16.json"

esi = ESIClientProvider(
    compatibility_date="2025-12-16",
    ua_appname="aa-standingssync",
    ua_version=__version__,
    operations=[
        "DeleteCharactersCharacterIdContacts",
        "GetAlliancesAllianceIdContacts",
        "GetCharactersCharacterIdContacts",
        "GetCharactersCharacterIdContactsLabels",
        "GetCorporationsCorporationId",
        "GetWars",
        "GetWarsWarId",
        "PostCharactersAffiliation",
        "PostCharactersCharacterIdContacts",
        "PutCharactersCharacterIdContacts",
    ],
    spec_file=spec_file,
)



import datetime as dt
from http import HTTPStatus
from unittest.mock import patch

import pook

from django.test import TestCase, override_settings
from django.utils.timezone import now
from eveuniverse.tests.testdata.factories_2 import (
    EveEntityAllianceFactory,
    EveEntityCharacterFactory,
)

from allianceauth.eveonline.models import EveAllianceInfo

from standingssync import tasks
from standingssync.models import EveWar
from standingssync.tests.factories import (
    EveWarFactory,
    SyncedCharacterFactory,
    SyncManagerFactory,
    UserMainDefaultFactory,
    UserMainManagerFactory,
    make_esi_url,
)
from standingssync.tests.helpers import TestCaseWithClearCache, extract

ESI_CONTACTS_PATH = "standingssync.core.esi_contacts"
ESI_API_PATH = "standingssync.core.esi_api"
MODELS_PATH = "standingssync.models"


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestTasksE2E(TestCaseWithClearCache):
    @pook.on
    def test_should_sync_manager_and_character_without_war_targets_and_compress(self):
        # given
        user_1 = UserMainManagerFactory()
        sm = SyncManagerFactory(user=user_1)
        alliance_id = user_1.profile.main_character.alliance_id
        contact_1_id = 1001
        EveEntityCharacterFactory(id=contact_1_id)
        standing = 9.9
        pook.get(
            make_esi_url(f"alliances/{alliance_id}/contacts"),
            reply=HTTPStatus.OK,
            response_headers={"X-Pages": "1"},
            response_json=[
                {
                    "contact_id": contact_1_id,
                    "contact_type": "character",
                    "standing": standing,
                }
            ],
        )
        sc = SyncedCharacterFactory(manager=sm)
        pook.get(
            make_esi_url(f"characters/{sc.character_id}/contacts"),
            reply=HTTPStatus.OK,
            response_headers={"X-Pages": "1"},
            response_json=[],
        )
        pook.get(
            make_esi_url(f"characters/{sc.character_id}/contacts/labels"),
            reply=HTTPStatus.OK,
            response_headers={"X-Pages": "1"},
            response_json=[],
        )
        pook.post(
            url=make_esi_url(f"characters/{sc.character_id}/contacts"),
            params={"standing": str(10.0)},
            json=[sm.alliance.alliance_id],
            reply=HTTPStatus.CREATED,
            response_json=[sm.alliance.id],
        )
        pook.post(
            url=make_esi_url(f"characters/{sc.character_id}/contacts"),
            params={"standing": str(standing)},
            json=[contact_1_id],
            reply=HTTPStatus.CREATED,
            response_json=[contact_1_id],
        )

        # when
        with (
            patch(MODELS_PATH + ".STANDINGSSYNC_REPLACE_CONTACTS", True),
            patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", False),
        ):
            tasks.run_manager_sync.delay(manager_pk=sm.pk)

        # then
        self.assertTrue(pook.isdone())

    @pook.on
    def test_should_sync_manager_and_character_with_war_targets_and_no_compress(self):
        # given
        user_1 = UserMainManagerFactory()
        sm = SyncManagerFactory(user=user_1)
        alliance_id = user_1.profile.main_character.alliance_id
        contact_1_id = 1001
        EveEntityCharacterFactory(id=contact_1_id)
        standing = 9.9
        aggressor = EveEntityAllianceFactory()
        EveWarFactory(
            aggressor=aggressor, defender=EveEntityAllianceFactory(id=alliance_id)
        )
        pook.get(
            make_esi_url(f"alliances/{alliance_id}/contacts"),
            reply=HTTPStatus.OK,
            response_headers={"X-Pages": "1"},
            response_json=[
                {
                    "contact_id": contact_1_id,
                    "contact_type": "character",
                    "standing": standing,
                }
            ],
        )
        sc = SyncedCharacterFactory(manager=sm)
        pook.get(
            make_esi_url(f"characters/{sc.character_id}/contacts"),
            reply=HTTPStatus.OK,
            response_headers={"X-Pages": "1"},
            response_json=[],
        )
        wt_label_name = "WAR TARGETS"
        wt_label_id = 7
        pook.get(
            make_esi_url(f"characters/{sc.character_id}/contacts/labels"),
            reply=HTTPStatus.OK,
            response_headers={"X-Pages": "1"},
            response_json=[
                {
                    "label_id": wt_label_id,
                    "label_name": wt_label_name,
                }
            ],
        )
        pook.post(
            url=make_esi_url(f"characters/{sc.character_id}/contacts"),
            params={"standing": str(10.0)},
            json=[sm.alliance.alliance_id],
            reply=HTTPStatus.CREATED,
            response_json=[sm.alliance.id],
        )
        pook.post(
            url=make_esi_url(f"characters/{sc.character_id}/contacts"),
            params={"standing": str(standing)},
            json=[contact_1_id],
            reply=HTTPStatus.CREATED,
            response_json=[contact_1_id],
        )
        pook.post(
            url=make_esi_url(f"characters/{sc.character_id}/contacts"),
            params={"standing": str(-10.0), "label_ids": [str(wt_label_id)]},
            json=[aggressor.id],
            reply=HTTPStatus.CREATED,
            response_json=[aggressor.id],
        )

        # when
        with (
            patch(MODELS_PATH + ".STANDINGSSYNC_REPLACE_CONTACTS", True),
            patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True),
            patch(
                ESI_CONTACTS_PATH + ".STANDINGSSYNC_WAR_TARGETS_LABEL_NAME",
                wt_label_name,
            ),
        ):
            tasks.run_manager_sync.delay(manager_pk=sm.pk)

        # then
        self.assertTrue(pook.isdone())

    @pook.on
    def test_should_sync_manager_and_character_with_war_targets_and_compress(self):
        # given
        user_1 = UserMainManagerFactory()
        sm = SyncManagerFactory(user=user_1, compress_contacts=True)
        alliance_id = user_1.profile.main_character.alliance_id
        contact_1_id = 1001
        EveEntityCharacterFactory(id=contact_1_id)
        standing = 9.9
        aggressor = EveEntityAllianceFactory()
        EveWarFactory(
            aggressor=aggressor, defender=EveEntityAllianceFactory(id=alliance_id)
        )
        pook.post(
            make_esi_url("characters/affiliation"),
            reply=HTTPStatus.OK,
            response_json=[
                {
                    "character_id": contact_1_id,
                    "corporation_id": 2011,
                },
            ],
        )
        pook.get(
            make_esi_url(f"alliances/{alliance_id}/contacts"),
            reply=HTTPStatus.OK,
            response_headers={"X-Pages": "1"},
            response_json=[
                {
                    "contact_id": contact_1_id,
                    "contact_type": "character",
                    "standing": standing,
                }
            ],
        )
        sc = SyncedCharacterFactory(manager=sm)
        pook.get(
            make_esi_url(f"characters/{sc.character_id}/contacts"),
            reply=HTTPStatus.OK,
            response_headers={"X-Pages": "1"},
            response_json=[],
        )
        wt_label_name = "WAR TARGETS"
        wt_label_id = 7
        pook.get(
            make_esi_url(f"characters/{sc.character_id}/contacts/labels"),
            reply=HTTPStatus.OK,
            response_headers={"X-Pages": "1"},
            response_json=[
                {
                    "label_id": wt_label_id,
                    "label_name": wt_label_name,
                }
            ],
        )
        pook.post(
            url=make_esi_url(f"characters/{sc.character_id}/contacts"),
            params={"standing": str(10.0)},
            json=[sm.alliance.alliance_id],
            reply=HTTPStatus.CREATED,
            response_json=[sm.alliance.id],
        )
        pook.post(
            url=make_esi_url(f"characters/{sc.character_id}/contacts"),
            params={"standing": str(standing)},
            json=[contact_1_id],
            reply=HTTPStatus.CREATED,
            response_json=[contact_1_id],
        )
        pook.post(
            url=make_esi_url(f"characters/{sc.character_id}/contacts"),
            params={"standing": str(-10.0), "label_ids": [str(wt_label_id)]},
            json=[aggressor.id],
            reply=HTTPStatus.CREATED,
            response_json=[aggressor.id],
        )

        # when
        with (
            patch(MODELS_PATH + ".STANDINGSSYNC_REPLACE_CONTACTS", True),
            patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True),
            patch(
                ESI_CONTACTS_PATH + ".STANDINGSSYNC_WAR_TARGETS_LABEL_NAME",
                wt_label_name,
            ),
        ):
            tasks.run_manager_sync.delay(manager_pk=sm.pk)

        # then
        self.assertTrue(pook.isdone())

    @pook.on
    def test_sync_wars(self):
        # given
        war_id = 719980
        pook.get(
            make_esi_url("wars"),
            reply=HTTPStatus.OK,
            response_json=[war_id],
        )
        pook.get(
            make_esi_url(f"wars/{war_id}"),
            reply=HTTPStatus.OK,
            response_json={
                "aggressor": {
                    "alliance_id": EveEntityAllianceFactory().id,
                    "isk_destroyed": 0,
                    "ships_killed": 0,
                },
                "declared": (now() - dt.timedelta(days=5)).isoformat(),
                "defender": {
                    "alliance_id": EveEntityAllianceFactory().id,
                    "isk_destroyed": 0,
                    "ships_killed": 0,
                },
                "id": war_id,
                "mutual": False,
                "open_for_allies": True,
                "started": (now() - dt.timedelta(days=4)).isoformat(),
            },
        )

        # when
        with (
            patch(ESI_API_PATH + ".STANDINGSSYNC_UNFINISHED_WARS_EXCEPTION_IDS", []),
            patch(ESI_API_PATH + ".STANDINGSSYNC_UNFINISHED_WARS_MINIMUM_ID", 0),
        ):
            tasks.sync_all_wars.delay()

        # then
        got = extract(EveWar.objects, "id")
        want = {war_id}
        self.assertEqual(got, want)


class TestUI(TestCase):
    def test_should_open_main_page_wo_syn_manager(self):
        # given
        user = UserMainDefaultFactory()
        self.client.force_login(user)
        # when
        response = self.client.get("/standingssync/characters")
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_should_open_main_page_w_sync_manager_and_chars(self):
        # given
        user = UserMainDefaultFactory()
        alliance = EveAllianceInfo.objects.get(
            alliance_id=user.profile.main_character.alliance_id
        )
        sync_manager = SyncManagerFactory(alliance=alliance)
        SyncedCharacterFactory(manager=sync_manager)
        self.client.force_login(user)
        # when
        response = self.client.get("/standingssync/characters")
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_should_open_wars_page_w_sync_manager(self):
        # given
        user = UserMainDefaultFactory()
        alliance = EveAllianceInfo.objects.get(
            alliance_id=user.profile.main_character.alliance_id
        )
        sync_manager = SyncManagerFactory(alliance=alliance)
        SyncedCharacterFactory(manager=sync_manager)
        self.client.force_login(user)
        # when
        response = self.client.get("/standingssync/wars")
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_should_open_wars_page_wo_sync_manager(self):
        # given
        user = UserMainDefaultFactory()
        self.client.force_login(user)
        # when
        response = self.client.get("/standingssync/wars")
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)



"""Helpers for standingssync."""

import json
from pathlib import Path


def store_json(data, filename: str) -> None:
    """Store data as JSON in a file."""
    path = Path.cwd() / f"{filename}.json"
    with path.open("w", encoding="utf-8") as file:
        json.dump(data, file, indent=4)



"""Wrapper for handling all access to the ESI API."""

from collections import defaultdict
from typing import Callable, Dict, FrozenSet, Iterable, Optional, Set

from esi.models import Token

from allianceauth.services.hooks import get_extension_logger
from app_utils.helpers import chunks

from standingssync.app_settings import (
    STANDINGSSYNC_UNFINISHED_WARS_EXCEPTION_IDS,
    STANDINGSSYNC_UNFINISHED_WARS_MINIMUM_ID,
)
from standingssync.core.esi_contacts import EsiContact, EsiContactLabel
from standingssync.providers import esi

logger = get_extension_logger(__name__)

FETCH_WARS_MAX_ITEMS = 2000

logger = get_extension_logger(__name__)


def add_character_contacts(token: Token, contacts: Iterable[EsiContact]) -> None:
    """Add new contacts on ESI for a character."""
    _update_character_contacts(
        token=token,
        contacts=contacts,
        esi_method=esi.client.Contacts.PostCharactersCharacterIdContacts,
    )
    logger.info("%s: Added %d contacts", token.character_name, len(contacts))


def delete_character_contacts(token: Token, contacts: Iterable[EsiContact]):
    """Delete character contacts on ESI."""
    max_items = 20
    contact_ids = sorted([contact.contact_id for contact in contacts])
    contact_ids_chunks = chunks(contact_ids, max_items)
    for contact_ids_chunk in contact_ids_chunks:
        esi.client.Contacts.DeleteCharactersCharacterIdContacts(
            token=token,
            character_id=token.character_id,
            contact_ids=contact_ids_chunk,
        ).result(use_etag=False)

    logger.info("%s: Deleted %d contacts", token.character_name, len(contact_ids))


def fetch_alliance_contacts(alliance_id: int, token: Token) -> Set[EsiContact]:
    """Fetch alliance contacts from ESI."""
    contacts_raw = esi.client.Contacts.GetAlliancesAllianceIdContacts(
        token=token, alliance_id=alliance_id
    ).results(use_etag=False)
    contacts = {
        row.contact_id: EsiContact.from_esi_dict(row.model_dump())
        for row in contacts_raw
    }
    # add the sync alliance with max standing to contacts
    contacts[alliance_id] = EsiContact(
        contact_id=alliance_id,
        contact_type=EsiContact.Category.ALLIANCE,
        standing=10,
    )
    return set(contacts.values())


def fetch_character_contacts(token: Token) -> Set[EsiContact]:
    """Fetch character contacts from ESI."""
    character_contacts_raw = esi.client.Contacts.GetCharactersCharacterIdContacts(
        token=token, character_id=token.character_id
    ).results(use_etag=False)
    logger.info(
        "%s: Fetched %d current contacts",
        token.character_name,
        len(character_contacts_raw),
    )
    character_contacts = {
        EsiContact.from_esi_dict(contact.model_dump())
        for contact in character_contacts_raw
    }
    return character_contacts


def fetch_character_contact_labels(token: Token) -> Set[EsiContactLabel]:
    """Fetch contact labels for character from ESI."""
    labels_raw = esi.client.Contacts.GetCharactersCharacterIdContactsLabels(
        character_id=token.character_id, token=token
    ).result(use_etag=False)
    logger.info("%s: Fetched %d current labels", token.character_name, len(labels_raw))
    labels = {EsiContactLabel.from_esi_dict(label.model_dump()) for label in labels_raw}
    return labels


def update_character_contacts(token: Token, contacts: Iterable[EsiContact]) -> None:
    """Update existing character contacts on ESI."""
    _update_character_contacts(
        token=token,
        contacts=contacts,
        esi_method=esi.client.Contacts.PutCharactersCharacterIdContacts,
    )
    logger.info("%s: Updated %d contacts", token.character_name, len(contacts))


def _update_character_contacts(
    token: Token, contacts: Iterable[EsiContact], esi_method: Callable
) -> None:
    for label_ids, contacts_by_standing in _group_for_esi_update(contacts).items():
        _update_character_contacts_esi(
            token=token,
            contacts_by_standing=contacts_by_standing,
            esi_method=esi_method,
            label_ids=list(label_ids) if label_ids else None,
        )


def _update_character_contacts_esi(
    token: Token,
    contacts_by_standing: Dict[float, Iterable[int]],
    esi_method: Callable,
    label_ids: Optional[list] = None,
) -> None:
    """Add new or update existing character contacts on ESI."""
    max_items = 100
    for standing in contacts_by_standing:
        contact_ids = sorted(list(contacts_by_standing[standing]))
        for contact_ids_chunk in chunks(contact_ids, max_items):
            params = {
                "token": token,
                "character_id": token.character_id,
                "body": contact_ids_chunk,
                "standing": standing,
            }
            if label_ids is not None:
                params["label_ids"] = sorted(list(label_ids))
            esi_method(**params).result(use_etag=False)


def _group_for_esi_update(
    contacts: Iterable["EsiContact"],
) -> Dict[FrozenSet, Dict[float, Iterable[int]]]:
    """Group contacts for ESI update."""
    contacts_grouped = {}
    for contact in contacts:
        if contact.label_ids not in contacts_grouped:
            contacts_grouped[contact.label_ids] = defaultdict(set)
        contacts_grouped[contact.label_ids][contact.standing].add(contact.contact_id)
    return contacts_grouped


def fetch_war_ids(min_war_id: int = 0) -> Set[int]:
    """Fetch IDs for new and unfinished wars from ESI.

    Will ignore older wars which are known to be already finished.

    Args:
        min_war_id: when provided will only return war IDs higher then this value.
            this prevents this function to re-fetch the same war IDs again from ESI.
    """
    war_ids = []
    war_ids_page = esi.client.Wars.GetWars().result(use_etag=False)

    while True:
        war_ids += war_ids_page
        if (
            len(war_ids_page) < FETCH_WARS_MAX_ITEMS
            or min(war_ids_page) < STANDINGSSYNC_UNFINISHED_WARS_MINIMUM_ID
            or min(war_ids_page) <= min_war_id
        ):
            break

        max_war_id = min(war_ids)
        war_ids_page = esi.client.Wars.GetWars(max_war_id=max_war_id).result(
            use_etag=False
        )

    logger.info("Fetched %d war IDs from ESI", len(war_ids))

    war_ids = {
        war_id
        for war_id in war_ids
        if war_id >= STANDINGSSYNC_UNFINISHED_WARS_MINIMUM_ID
    }
    war_ids = war_ids.union(set(STANDINGSSYNC_UNFINISHED_WARS_EXCEPTION_IDS))
    if min_war_id:
        war_ids = {war_id for war_id in war_ids if war_id > min_war_id}

    return war_ids


def fetch_war(war_id: int) -> dict:
    """Fetch details about a war from ESI."""
    war_info = esi.client.Wars.GetWarsWarId(war_id=war_id).result(use_etag=False)
    logger.info("Retrieved war details for ID %s", war_id)
    return war_info.model_dump()



"""Settings for standingssync."""

from app_utils.app_settings import clean_setting

STANDINGSSYNC_ADD_WAR_TARGETS = clean_setting("STANDINGSSYNC_ADD_WAR_TARGETS", False)
"""When enabled will automatically add or set war targets
 with standing = -10 to synced characters.
 Note that for this to work a character also needs to have created
 a custom label with the correct name.
 See also `STANDINGSSYNC_WAR_TARGETS_LABEL_NAME`.
"""

STANDINGSSYNC_CHAR_MIN_STANDING = clean_setting(
    "STANDINGSSYNC_CHAR_MIN_STANDING", default_value=0.1, min_value=-10, max_value=10
)
"""Minimum standing a character needs to have in order to get alliance contacts.
Any char with a standing smaller than this value will be rejected.
Set to `0.0` if you want to allow neutral alts to sync.
"""

STANDINGSSYNC_STORE_ESI_CONTACTS_ENABLED = clean_setting(
    "STANDINGSSYNC_STORE_ESI_CONTACTS_ENABLED", False
)
"""Wether to store contacts received from ESI to disk. This is for debugging."""

STANDINGSSYNC_REPLACE_CONTACTS = clean_setting("STANDINGSSYNC_REPLACE_CONTACTS", True)
"""When enabled will replace contacts of synced characters with alliance contacts."""

STANDINGSSYNC_SYNC_TIMEOUT = clean_setting("STANDINGSSYNC_SYNC_TIMEOUT", 180)  # 3 hours
"""Duration in minutes after which a delayed sync for managers and characters
is reported as down. This value should be aligned with the frequency of the sync task.
"""

STANDINGSSYNC_UNFINISHED_WARS_EXCEPTION_IDS = clean_setting(
    "STANDINGSSYNC_UNFINISHED_WARS_EXCEPTION_IDS",
    [
        734022,
        748895,
        756575,
        757763,
        758152,
        758264,
        758295,
        758365,
        758370,
        758447,
        758739,
        758761,
        758770,
        758945,
        758969,
        758970,
        758971,
        759161,
        759533,
        759534,
        759535,
        759536,
        759581,
        759582,
        759584,
        759585,
        759586,
        759587,
        759592,
        759593,
    ],
)
"""IDs of unfinished wars, with IDs below the above minimum threshold.
These wars will also be included in the list of unfinished wars,
despite there ID being below the minimum ID.

:private:
"""

STANDINGSSYNC_UNFINISHED_WARS_MINIMUM_ID = clean_setting(
    "STANDINGSSYNC_UNFINISHED_WARS_MINIMUM_ID", 759665
)
"""Smallest war ID to fetch from ESI.

All wars with smaller IDs are known to be already finished.
This is an optimization to avoid having to fetch >700K wars from ESI.

:private:
"""

STANDINGSSYNC_WAR_TARGETS_LABEL_NAME = clean_setting(
    "STANDINGSSYNC_WAR_TARGETS_LABEL_NAME", "WAR TARGETS"
)
"""Name of EVE contact label for war targets.
Needs to be created by the user for each synced character. Required to ensure that
war targets are deleted once they become invalid. Not case sensitive.
"""



from http import HTTPStatus
from unittest.mock import patch

import pook

from eveuniverse.tests.testdata.factories_2 import EveEntityCharacterFactory

from app_utils.testing import NoSocketsTestCase

from standingssync.core import esi_api
from standingssync.core.esi_contacts import EsiContact
from standingssync.tests.factories import (
    EsiContactCharacterFactory,
    EsiContactLabelFactory,
    UserMainDefaultFactory,
    UserMainManagerFactory,
    make_esi_url,
)
from standingssync.tests.helpers import TestCaseWithClearCache

MODULE_PATH = "standingssync.core.esi_api"


class TestEsiApi(TestCaseWithClearCache):
    @pook.on
    def test_should_fetch_alliance_contacts(self):
        # given
        user = UserMainManagerFactory()
        alliance_id = user.profile.main_character.alliance_id
        token = user.token_set.first()
        contact_id = 1001
        EveEntityCharacterFactory
        pook.get(
            make_esi_url(f"alliances/{alliance_id}/contacts"),
            reply=HTTPStatus.OK,
            response_headers={"X-Pages": "1"},
            response_json=[
                {
                    "contact_id": contact_id,
                    "contact_type": "character",
                    "standing": 9.9,
                }
            ],
        )

        # when
        result = esi_api.fetch_alliance_contacts(alliance_id=alliance_id, token=token)

        # then
        expected = {
            EsiContact(contact_id, EsiContact.Category.CHARACTER, 9.9),
            EsiContact(alliance_id, EsiContact.Category.ALLIANCE, 10),
        }
        self.assertSetEqual(expected, result)

    @pook.on
    def test_should_fetch_character_contacts(self):
        # given
        user = UserMainDefaultFactory()
        character_id = user.profile.main_character.character_id
        token = user.token_set.first()
        contact_id = 1001
        pook.get(
            make_esi_url(f"characters/{character_id}/contacts"),
            reply=HTTPStatus.OK,
            response_headers={"X-Pages": "1"},
            response_json=[
                {
                    "contact_id": contact_id,
                    "contact_type": "corporation",
                    "standing": 9.9,
                }
            ],
        )

        # when
        result = esi_api.fetch_character_contacts(token=token)

        # then
        expected = {EsiContact(contact_id, EsiContact.Category.CORPORATION, 9.9)}
        self.assertSetEqual(expected, result)

    @pook.on
    def test_should_fetch_contact_labels(self):
        # given
        user = UserMainDefaultFactory()
        character_id = user.profile.main_character.character_id
        token = user.token_set.first()
        label_1 = EsiContactLabelFactory()
        label_2 = EsiContactLabelFactory()
        pook.get(
            make_esi_url(f"characters/{character_id}/contacts/labels"),
            reply=HTTPStatus.OK,
            response_json=[
                label_1.to_esi_dict(),
                label_2.to_esi_dict(),
            ],
        )

        # when
        result = esi_api.fetch_character_contact_labels(token=token)

        # then
        expected = {label_1, label_2}
        self.assertSetEqual(result, expected)

    @pook.on
    def test_should_add_character_contact(self):
        # given
        user = UserMainDefaultFactory()
        character_id = user.profile.main_character.character_id
        token = user.token_set.first()
        standing = 5.0
        contact = EsiContact.from_eve_entity(
            EveEntityCharacterFactory(), standing=standing
        )
        pook.post(
            url=make_esi_url(f"characters/{character_id}/contacts"),
            params={"standing": str(standing)},
            json=[contact.contact_id],
            reply=HTTPStatus.CREATED,
            response_json=[contact.contact_id],
        )

        # when
        esi_api.add_character_contacts(token, {contact})

        # then
        self.assertTrue(pook.isdone())

    @pook.on
    def test_should_update_character_contact(self):
        # given
        user = UserMainDefaultFactory()
        character_id = user.profile.main_character.character_id
        token = user.token_set.first()
        standing = 5.0
        contact = EsiContact.from_eve_entity(
            EveEntityCharacterFactory(), standing=standing
        )
        pook.put(
            url=make_esi_url(f"characters/{character_id}/contacts"),
            params={"standing": str(standing)},
            json=[contact.contact_id],
            reply=HTTPStatus.NO_CONTENT,
        )

        # when
        esi_api.update_character_contacts(token, {contact})

        # then
        self.assertTrue(pook.isdone())

    @pook.on
    def test_should_delete_character_contact(self):
        # given
        user = UserMainDefaultFactory()
        character_id = user.profile.main_character.character_id
        token = user.token_set.first()
        contact = EsiContact.from_eve_entity(EveEntityCharacterFactory(), standing=5.0)
        pook.delete(
            url=make_esi_url(f"characters/{character_id}/contacts"),
            params={"contact_ids": [str(contact.contact_id)]},
            reply=HTTPStatus.NO_CONTENT,
        )

        # when
        esi_api.delete_character_contacts(token, {contact})

        # then
        self.assertTrue(pook.isdone())


class TestEsiContactsHelpers(NoSocketsTestCase):
    def test_should_group_contacts_for_esi_update(self):
        # given
        label_1 = EsiContactLabelFactory(id=1)
        contact_1 = EsiContactCharacterFactory(contact_id=11, label_ids=[label_1.id])
        label_2 = EsiContactLabelFactory(id=2)
        contact_2 = EsiContactCharacterFactory(
            contact_id=12, label_ids=[label_1.id, label_2.id]
        )
        contact_3 = EsiContactCharacterFactory(contact_id=13, standing=2.0)
        contact_4 = EsiContactCharacterFactory(contact_id=14, standing=2.0)
        esi_contacts = [contact_1, contact_2, contact_3, contact_4]
        # when
        result = esi_api._group_for_esi_update(esi_contacts)
        self.maxDiff = None
        # then
        expected = {
            frozenset({1}): {contact_1.standing: {contact_1.contact_id}},
            frozenset({1, 2}): {contact_2.standing: {contact_2.contact_id}},
            frozenset(): {2.0: {contact_3.contact_id, contact_4.contact_id}},
        }
        self.assertEqual(expected, result)


class TestEsiWarsIDs(TestCaseWithClearCache):
    @pook.on
    def test_should_fetch_war_ids(self):
        # given
        exception_ids = [1]
        minimum_id = 3
        war_ids = [5, 4, 3, 2]
        pook.get(
            make_esi_url("wars"),
            reply=HTTPStatus.OK,
            response_json=war_ids,
        )

        # when
        with (
            patch(
                MODULE_PATH + ".STANDINGSSYNC_UNFINISHED_WARS_EXCEPTION_IDS",
                exception_ids,
            ),
            patch(
                MODULE_PATH + ".STANDINGSSYNC_UNFINISHED_WARS_MINIMUM_ID", minimum_id
            ),
        ):
            got = esi_api.fetch_war_ids()

        # then
        self.assertSetEqual(got, {5, 4, 3, 1})
        self.assertTrue(pook.isdone())

    @pook.on
    def test_should_fetch_unknown_war_ids_only(self):
        # given
        exception_ids = [1]
        minimum_id = 3
        war_ids = [5, 4, 3, 2]
        pook.get(
            make_esi_url("wars"),
            reply=HTTPStatus.OK,
            response_json=war_ids,
        )

        # when
        with (
            patch(
                MODULE_PATH + ".STANDINGSSYNC_UNFINISHED_WARS_EXCEPTION_IDS",
                exception_ids,
            ),
            patch(
                MODULE_PATH + ".STANDINGSSYNC_UNFINISHED_WARS_MINIMUM_ID", minimum_id
            ),
        ):
            got = esi_api.fetch_war_ids(4)

        # then
        self.assertSetEqual(got, {5})
        self.assertTrue(pook.isdone())

    @pook.on
    def test_should_fetch_war_ids_with_paging(self):
        # given
        exception_ids = [2]
        minimum_id = 4
        war_ids = [6, 5, 4]
        page_size = 2
        pook.get(
            make_esi_url("wars"),
            reply=HTTPStatus.OK,
            response_json=war_ids[:page_size],
        )
        pook.get(
            make_esi_url("wars"),
            params={"max_war_id": str(5)},
            reply=HTTPStatus.OK,
            response_json=war_ids[page_size:],
        )

        # when
        with (
            patch(
                MODULE_PATH + ".STANDINGSSYNC_UNFINISHED_WARS_EXCEPTION_IDS",
                exception_ids,
            ),
            patch(
                MODULE_PATH + ".STANDINGSSYNC_UNFINISHED_WARS_MINIMUM_ID", minimum_id
            ),
            patch(MODULE_PATH + ".FETCH_WARS_MAX_ITEMS", page_size),
        ):
            got = esi_api.fetch_war_ids()

        # then
        self.assertSetEqual(got, set(war_ids) | set(exception_ids))
        self.assertTrue(pook.isdone())



import datetime as dt
from unittest.mock import MagicMock, patch

from django.utils.timezone import now
from eveuniverse.tests.testdata.factories_2 import EveEntityAllianceFactory

from app_utils.testdata_factories import EveAllianceInfoFactory, UserFactory
from app_utils.testing import NoSocketsTestCase

from standingssync.models import EveWar, SyncManager
from standingssync.tests.factories import (
    EveContactFactory,
    EveWarEmptyFactory,
    EveWarFactory,
    SyncManagerFactory,
    UserMainDefaultFactory,
)
from standingssync.tests.helpers import extract

MANAGERS_PATH = "standingssync.managers"


class TestEveContactManager(NoSocketsTestCase):
    def test_grouped_by_standing(self):
        # given
        sync_manager = SyncManagerFactory()

        contact_terrible = EveContactFactory(
            manager=sync_manager,
            standing=-10.0,
        )
        contact_bad = EveContactFactory(
            manager=sync_manager,
            standing=-5.0,
        )
        contact_neutral = EveContactFactory(
            manager=sync_manager,
            standing=0.0,
        )
        contact_good = EveContactFactory(
            manager=sync_manager,
            standing=5.0,
        )
        contact_excellent = EveContactFactory(
            manager=sync_manager,
            standing=10.0,
        )

        expected = {
            -10.0: {contact_terrible},
            -5.0: {contact_bad},
            0.0: {contact_neutral},
            5.0: {contact_good},
            10.0: {contact_excellent},
        }

        # when
        result = sync_manager.contacts.grouped_by_standing()

        # then
        self.maxDiff = None
        self.assertDictEqual(result, expected)
        self.assertListEqual(list(result.keys()), list(expected.keys()))


class TestEveWarManagerWarTargets(NoSocketsTestCase):
    def test_should_return_defender_and_allies_for_aggressor(self):
        # given
        aggressor = EveEntityAllianceFactory()
        defender = EveEntityAllianceFactory()
        ally_1 = EveEntityAllianceFactory()
        ally_2 = EveEntityAllianceFactory()
        EveWarFactory(aggressor=aggressor, defender=defender, allies=[ally_1, ally_2])
        alliance = EveAllianceInfoFactory(alliance_id=aggressor.id)
        # when
        result = EveWar.objects.alliance_war_targets(alliance)
        # then
        self.assertSetEqual(
            {obj.id for obj in result}, {defender.id, ally_1.id, ally_2.id}
        )

    def test_should_return_aggressor_for_defender(self):
        # given
        aggressor = EveEntityAllianceFactory()
        defender = EveEntityAllianceFactory()
        ally = EveEntityAllianceFactory()
        EveWarFactory(aggressor=aggressor, defender=defender, allies=[ally])
        alliance = EveAllianceInfoFactory(alliance_id=defender.id)
        # when
        result = EveWar.objects.alliance_war_targets(alliance)
        # then
        self.assertSetEqual({obj.id for obj in result}, {aggressor.id})

    def test_should_return_aggressor_for_ally(self):
        # given
        aggressor = EveEntityAllianceFactory()
        defender = EveEntityAllianceFactory()
        ally = EveEntityAllianceFactory()
        EveWarFactory(aggressor=aggressor, defender=defender, allies=[ally])
        alliance = EveAllianceInfoFactory(alliance_id=ally.id)
        # when
        result = EveWar.objects.alliance_war_targets(alliance)
        # then
        self.assertSetEqual({obj.id for obj in result}, {aggressor.id})


class TestEveWarQueryset(NoSocketsTestCase):
    def test_should_return_wars_of_alliance_only(self):
        # given
        alliance_entity = EveEntityAllianceFactory()
        alliance = EveAllianceInfoFactory(alliance_id=alliance_entity.id)
        other_1 = EveEntityAllianceFactory()
        other_2 = EveEntityAllianceFactory()
        war_1 = EveWarFactory(aggressor=alliance_entity, defender=other_1)
        war_2 = EveWarFactory(aggressor=other_1, defender=alliance_entity)
        war_3 = EveWarFactory(
            aggressor=other_1, defender=other_2, allies=[alliance_entity]
        )
        EveWarFactory(aggressor=other_1, defender=other_2)

        # when
        qs = EveWar.objects.alliance_wars(alliance)

        # then
        expected = {war_1.id, war_2.id, war_3.id}
        result = set(qs.values_list("id", flat=True))
        self.assertSetEqual(expected, result)

    def test_should_return_non_empty_wars(self):
        # given
        war = EveWarFactory()
        EveWarEmptyFactory()

        # when
        got = EveWar.objects.non_empty()

        # then
        self.assertCountEqual(got, [war])

    def test_should_return_empty_wars(self):
        # given
        EveWarFactory()
        war = EveWarEmptyFactory()

        # when
        got = EveWar.objects.empty()

        # then
        self.assertCountEqual(got, [war])

    def test_should_return_wars_that_need_update_only(self):
        # given
        EveWarFactory()  # should not include recently synced active wars
        EveWarFactory(finished=now())  # should not include finished wars
        war_1 = EveWarEmptyFactory()  # should include empty wars
        with patch("django.utils.timezone.now") as m:
            m.return_value = now() - dt.timedelta(hours=1, seconds=1)
            war_2 = EveWarFactory()  # should include stale active wars
            EveWarFactory(finished=now())  # should not include stale finished wars

        # when
        got = EveWar.objects.needs_update()

        # then
        want = [war_1, war_2]
        self.assertCountEqual(got, want)


class TestEveWarManager_ActiveWars(NoSocketsTestCase):
    def test_should_return_started_war_as_defender(self):
        # given
        sync_manager = SyncManagerFactory()
        war = EveWarFactory(
            defender=EveEntityAllianceFactory(id=sync_manager.alliance.alliance_id),
            declared=now() - dt.timedelta(days=2),
        )
        # when
        result = EveWar.objects.active_wars()
        # then
        self.assertEqual(result.count(), 1)
        self.assertEqual(result.first(), war)

    def test_should_return_started_war_as_attacker(self):
        # given
        sync_manager = SyncManagerFactory()
        war = EveWarFactory(
            aggressor=EveEntityAllianceFactory(id=sync_manager.alliance.alliance_id),
            declared=now() - dt.timedelta(days=2),
        )
        # when
        result = EveWar.objects.active_wars()
        # then
        self.assertEqual(result.count(), 1)
        self.assertEqual(result.first(), war)

    def test_should_return_started_war_as_ally(self):
        # given
        sync_manager = SyncManagerFactory()
        war = EveWarFactory(
            allies=[EveEntityAllianceFactory(id=sync_manager.alliance.alliance_id)],
            declared=now() - dt.timedelta(days=2),
        )
        # when
        result = EveWar.objects.active_wars()
        # then
        self.assertEqual(result.count(), 1)
        self.assertEqual(result.first(), war)

    def test_should_return_war_about_to_finish(self):
        # given
        sync_manager = SyncManagerFactory()
        war = EveWarFactory(
            defender=EveEntityAllianceFactory(id=sync_manager.alliance.alliance_id),
            declared=now() - dt.timedelta(days=2),
            finished=now() + dt.timedelta(days=1),
        )
        # when
        result = EveWar.objects.active_wars()
        # then
        self.assertEqual(result.count(), 1)
        self.assertEqual(result.first(), war)

    def test_should_not_return_finished_war(self):
        # given
        sync_manager = SyncManagerFactory()
        EveWarFactory(
            defender=EveEntityAllianceFactory(id=sync_manager.alliance.alliance_id),
            declared=now() - dt.timedelta(days=2),
            finished=now() - dt.timedelta(days=1),
        )
        # when
        result = EveWar.objects.active_wars()
        # then
        self.assertEqual(result.count(), 0)

    def test_should_not_return_war_not_yet_started(self):
        # given
        sync_manager = SyncManagerFactory()
        EveWarFactory(
            defender=EveEntityAllianceFactory(id=sync_manager.alliance.alliance_id),
            declared=now() - dt.timedelta(days=1),
            started=now() + dt.timedelta(hours=4),
        )
        # when
        result = EveWar.objects.active_wars()
        # then
        self.assertEqual(result.count(), 0)


class TestEveWarManager_Annotations(NoSocketsTestCase):
    def test_should_annotate_state(self):
        # given
        war_pending = EveWarFactory(declared=now())
        war_ongoing = EveWarFactory(declared=now() - dt.timedelta(hours=24))
        war_concluding = EveWarFactory(finished=now() + dt.timedelta(hours=24))
        war_retracted = EveWarFactory(retracted=now())
        war_finished = EveWarFactory(finished=now() - dt.timedelta(hours=1))
        war_unknown = EveWarEmptyFactory()

        # when
        qs = EveWar.objects.annotate_state()

        # then
        self.assertEqual(qs.get(id=war_pending.id).state, EveWar.State.PENDING.value)
        self.assertEqual(qs.get(id=war_ongoing.id).state, EveWar.State.ONGOING.value)
        self.assertEqual(
            qs.get(id=war_concluding.id).state, EveWar.State.CONCLUDING.value
        )
        self.assertEqual(
            qs.get(id=war_retracted.id).state, EveWar.State.RETRACTED.value
        )
        self.assertEqual(qs.get(id=war_finished.id).state, EveWar.State.FINISHED.value)
        self.assertEqual(qs.get(id=war_unknown.id).state, EveWar.State.UNKNOWN.value)

    def test_should_annotate_is_active(self):
        # given
        war_pending = EveWarFactory(declared=now())
        war_ongoing = EveWarFactory(declared=now() - dt.timedelta(hours=24))
        war_concluding = EveWarFactory(finished=now() + dt.timedelta(hours=24))
        war_retracted = EveWarFactory(retracted=now())
        war_finished = EveWarFactory(finished=now() - dt.timedelta(hours=1))
        # when
        qs = EveWar.objects.annotate_state().annotate_is_active()
        # then
        self.assertFalse(qs.get(id=war_pending.id).is_active)
        self.assertTrue(qs.get(id=war_ongoing.id).is_active)
        self.assertTrue(qs.get(id=war_concluding.id).is_active)
        self.assertTrue(qs.get(id=war_retracted.id).is_active)
        self.assertFalse(qs.get(id=war_finished.id).is_active)


class TestEveWarManager_CurrentWars(NoSocketsTestCase):
    def test_should_return_recently_declared_war(self):
        # given
        war = EveWarFactory(declared=now())
        # when
        result = EveWar.objects.current_wars()
        # then
        self.assertEqual(result.count(), 1)
        self.assertEqual(result.first(), war)

    def test_should_return_recently_finished_war(self):
        # given
        war = EveWarFactory(finished=now() - dt.timedelta(hours=23))
        # when
        result = EveWar.objects.current_wars()
        # then
        self.assertEqual(result.count(), 1)
        self.assertEqual(result.first(), war)

    def test_should_return_active_war(self):
        # given
        war = EveWarFactory(declared=now() - dt.timedelta(days=2))
        # when
        result = EveWar.objects.current_wars()
        # then
        self.assertEqual(result.count(), 1)
        self.assertEqual(result.first(), war)


class TestEveWarManager_SyncKnownWars(NoSocketsTestCase):
    @patch(MANAGERS_PATH + ".esi_api.fetch_war_ids")
    def test_should_sync_wars(self, mock_fetch_war_ids_from_esi: MagicMock):
        # given
        mock_fetch_war_ids_from_esi.return_value = {99, 98, 42}
        EveWarFactory(id=42, finished=now() - dt.timedelta(days=1))
        # when
        EveWar.objects.sync_known_wars()
        # then
        args, _ = mock_fetch_war_ids_from_esi.call_args
        self.assertEqual(args[0], 42)
        got = extract(EveWar.objects, "id")
        want = {99, 98, 42}
        self.assertSetEqual(got, want)


class TestEveWarManager_UpdatedPercentage(NoSocketsTestCase):
    def test_should_return_correct_percentage(self):
        # given
        EveWarFactory()
        EveWarFactory()
        EveWarFactory()
        EveWarEmptyFactory()

        # when
        got = EveWar.objects.updated_percentage()

        # then
        self.assertEqual(got, 0.75)

    def test_should_return_1_when_completed(self):
        # given
        EveWarFactory()
        EveWarFactory()
        EveWarFactory()

        # when
        got = EveWar.objects.updated_percentage()

        # then
        self.assertEqual(got, 1)

    def test_should_return_0_when_no_wars(self):
        # when
        got = EveWar.objects.updated_percentage()

        # then
        self.assertEqual(got, 0)


class TestSyncManagerManager(NoSocketsTestCase):
    def test_should_return_matching_sync_manager(self):
        # given
        user = UserMainDefaultFactory()
        alliance = EveAllianceInfoFactory(
            alliance_id=user.profile.main_character.alliance_id
        )
        sync_manager = SyncManagerFactory(alliance=alliance)
        # when
        result = SyncManager.objects.fetch_for_user(user)
        # then
        self.assertEqual(result, sync_manager)

    def test_should_return_none_when_no_match(self):
        # given
        user = UserMainDefaultFactory()
        SyncManagerFactory()
        # when
        result = SyncManager.objects.fetch_for_user(user)
        # then
        self.assertIsNone(result)

    def test_should_return_none_when_user_has_no_main(self):
        # given
        user = UserFactory()
        # when
        result = SyncManager.objects.fetch_for_user(user)
        # then
        self.assertIsNone(result)



"""Managers for standingssync."""

# pylint: disable = redefined-builtin, missing-class-docstring

import datetime as dt
from collections import defaultdict
from typing import Any, Dict

from django.contrib.auth.models import User
from django.db import models
from django.db.models import Case, Max, Q, Value, When
from django.utils.timezone import now
from eveuniverse.models import EveEntity

from allianceauth.eveonline.models import EveAllianceInfo
from allianceauth.services.hooks import get_extension_logger

from standingssync.core import esi_api

logger = get_extension_logger(__name__)


class EveContactQuerySet(models.QuerySet):
    def grouped_by_standing(self) -> Dict[int, Any]:
        """Group alliance contacts by standing and convert into sorted dict."""
        contacts_by_standing = defaultdict(set)
        for contact in self.all():
            contacts_by_standing[contact.standing].add(contact)
        return dict(sorted(contacts_by_standing.items()))


class EveContactManagerBase(models.Manager):
    pass


EveContactManager = EveContactManagerBase.from_queryset(EveContactQuerySet)


class EveWarQuerySet(models.QuerySet):
    def annotate_state(self) -> models.QuerySet:
        """Add state field to queryset."""
        from standingssync.models import EveWar

        return self.annotate(
            state=Case(
                When(started__gt=now(), then=Value(EveWar.State.PENDING.value)),
                When(
                    started__lte=now(),
                    finished__isnull=True,
                    then=Value(EveWar.State.ONGOING.value),
                ),
                When(
                    started__lte=now(),
                    finished__gt=now(),
                    retracted__isnull=False,
                    then=Value(EveWar.State.RETRACTED.value),
                ),
                When(
                    started__lte=now(),
                    finished__gt=now(),
                    retracted__isnull=True,
                    then=Value(EveWar.State.CONCLUDING.value),
                ),
                When(
                    finished__isnull=False,
                    then=Value(EveWar.State.FINISHED.value),
                ),
                default=Value(EveWar.State.UNKNOWN.value),
            )
        )

    def annotate_is_active(self) -> models.QuerySet:
        """Add is_active field to queryset. Requires prior annotation of state."""
        return self.annotate(
            is_active=Case(
                When(state__in=self.model.State.active_states(), then=Value(True)),
                default=Value(False),
            )
        )

    def current_wars(self) -> models.QuerySet:
        """Add filter for current wars.

        This includes wars that are about to start,
        active wars and wars that ended recently.
        """
        cutoff = now() - dt.timedelta(hours=24)
        qs = self.filter(declared__lt=now())
        return (
            qs.filter(finished__gt=cutoff) | qs.filter(finished__isnull=True)
        ).distinct()

    def active_wars(self) -> models.QuerySet:
        """Add filter for active wars."""
        qs = self.filter(
            Q(started__lt=now()) & (Q(finished__gt=now()) | Q(finished__isnull=True))
        ).distinct()
        return qs

    def alliance_wars(self, alliance: EveAllianceInfo) -> models.QuerySet:
        """Include wars where a given alliance is participating only."""
        qs = (
            self.filter(
                Q(aggressor_id=alliance.alliance_id)
                | Q(defender_id=alliance.alliance_id)
                | Q(allies__id=alliance.alliance_id)
            )
        ).distinct()
        return qs

    def needs_update(self) -> models.QuerySet:
        """Filter for wars that need to be updated."""
        threshold = now() - dt.timedelta(hours=1)
        qs = self.filter(
            Q(finished__isnull=True)
            & (
                Q(aggressor__isnull=True)  # empty wars
                | Q(last_modified__lt=threshold)  # stale active war
            )
        ).distinct()
        return qs

    def non_empty(self) -> models.QuerySet:
        """Filter non-empty war objects."""
        return self.filter(
            aggressor__isnull=False, declared__isnull=False, defender__isnull=False
        )

    def empty(self) -> models.QuerySet:
        """Filter empty war objects."""
        return self.filter(
            Q(aggressor__isnull=True)
            | Q(declared__isnull=True)
            | Q(defender__isnull=True)
        )


class EveWarManagerBase(models.Manager):
    def alliance_war_targets(
        self, alliance: EveAllianceInfo
    ) -> models.QuerySet[EveEntity]:
        """Identify current war targets of on alliance."""
        war_target_ids = set()
        for war in self.alliance_wars(alliance).active_wars():
            # case 1 alliance is aggressor
            if war.aggressor_id == alliance.alliance_id:
                war_target_ids.add(war.defender_id)
                war_target_ids |= set(war.allies.values_list("id", flat=True))

            # case 2 alliance is defender
            if war.defender_id == alliance.alliance_id:
                war_target_ids.add(war.aggressor_id)

            # case 3 alliance is ally
            if war.allies.filter(id=alliance.alliance_id).exists():
                war_target_ids.add(war.aggressor_id)

        return EveEntity.objects.filter(id__in=war_target_ids)

    def sync_known_wars(self):
        """Synchronizes which wars are known. Wars are known if they have an ID."""
        from standingssync.models import EveWar

        min_war_id = self.aggregate(Max("id", default=0)).get("id__max") or 0
        war_ids = esi_api.fetch_war_ids(min_war_id)
        known_ids = set(self.values_list("id", flat=True))
        unknown_ids = war_ids - known_ids
        if not unknown_ids:
            logger.info("No new wars")
            return

        wars = (EveWar(id=war_id) for war_id in unknown_ids)
        EveWar.objects.bulk_create(wars, batch_size=500, ignore_conflicts=True)
        logger.info("Created %d new wars", len(unknown_ids))

    def updated_percentage(self) -> float:
        "Return the percentage of updated wars. Will return 0 when there are no wars."
        total = self.count()
        if not total:
            return 0
        updated = self.non_empty().count()
        return updated / total


EveWarManager = EveWarManagerBase.from_queryset(EveWarQuerySet)


class SyncManagerManager(models.Manager):
    def fetch_for_user(self, user: User) -> Any:
        """Fetch sync manager for given user. Return None if no match is found."""
        if not user.profile.main_character:
            return None

        try:
            alliance = EveAllianceInfo.objects.get(
                alliance_id=user.profile.main_character.alliance_id
            )
        except EveAllianceInfo.DoesNotExist:
            return None

        try:
            return self.get(alliance=alliance)
        except self.model.DoesNotExist:
            return None



from django.core.management.base import BaseCommand, CommandError
from django.db.models import QuerySet

from standingssync.models import EveWar


class Command(BaseCommand):
    help = (
        "Calculates the minimum and special war IDs "
        "to reduce how many wars need to be fetched from ESI."
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "--max-special-ids",
            default=20,
            type=int,
            help="Maximum number of special unfinished IDs to compile",
        )
        parser.add_argument(
            "--disable-checks",
            action="store_true",
            help="When enabled will check if local data is sufficient for calc",
        )

    def handle(self, *args, **options):
        wars: QuerySet[EveWar] = EveWar.objects.non_empty()
        total_wars = wars.count()
        if not options["disable_checks"]:
            if not total_wars:
                raise CommandError(
                    "No non-stale wars found in database. "
                    "Please update wars from ESI before running this command."
                )

            updated: float = EveWar.objects.updated_percentage()
            if updated < 1:
                raise CommandError(
                    f"Only {updated * 100:.0f} wars are updated. "
                    "Please finished updating all wars from ESI before running this command."
                )

        self.stdout.write(f"Wars in database: {total_wars}")

        special_unfinished_war_ids, min_unfinished_war_id = _calc_war_ids(
            options["max_special_ids"]
        )
        remaining_count = wars.filter(id__gte=min_unfinished_war_id).count() + len(
            special_unfinished_war_ids
        )

        self.stdout.write("Calculated new war ID values for settings are:")
        self.stdout.write(
            f"STANDINGSSYNC_UNFINISHED_WARS_MINIMUM_ID = {min_unfinished_war_id}"
        )
        self.stdout.write(
            f"STANDINGSSYNC_UNFINISHED_WARS_EXCEPTION_IDS = {special_unfinished_war_ids}"
        )
        self.stdout.write(f"Total wars to fetch: {remaining_count}")


def _calc_war_ids(max_special_ids: int):
    unfinished_war_ids = list(
        EveWar.objects.non_empty()
        .filter(finished__isnull=True)
        .order_by("id")
        .values_list("id", flat=True)
    )
    special_unfinished_war_ids = unfinished_war_ids[:max_special_ids]
    try:
        min_unfinished_war_id = unfinished_war_ids[max_special_ids]
    except IndexError:
        min_unfinished_war_id = max(special_unfinished_war_ids) + 1
    return special_unfinished_war_ids, min_unfinished_war_id



"""Utility functions and classes for tests"""

from dataclasses import dataclass
from typing import Iterable, Set
from unittest.mock import MagicMock

from django.core.cache import cache
from django.db.models import QuerySet
from django.test import TestCase
from esi.models import Token

from standingssync.core.esi_contacts import (
    EsiContact,
    EsiContactLabel,
    EsiContactsContainer,
)


@dataclass
class EsiCharacterContactsStub:
    """Simulates the contacts for a character on ESI"""

    _character_id: int
    _contacts: EsiContactsContainer = None

    def contacts(self) -> Set[EsiContact]:
        return self._contacts.contacts()

    def contact_ids(self) -> Set[int]:
        return {contact.contact_id for contact in self._contacts.contacts()}

    def _setup_esi_mock(self, mock_esi_api: MagicMock):
        """Sets the mock for ESI to this object."""
        mock_esi_api.add_character_contacts = self._add_contacts
        mock_esi_api.delete_character_contacts = self._delete_contacts
        mock_esi_api.fetch_character_contact_labels = self._fetch_labels
        mock_esi_api.fetch_character_contacts = self._fetch_contacts
        mock_esi_api.update_character_contacts = self._update_contacts

    def _setup_contacts(self, contacts):
        for contact in contacts:
            self._contacts.add_contact(contact)

    def _setup_labels(self, labels):
        for label in labels:
            self._contacts.add_label(label)

    @classmethod
    def create(
        cls,
        character_id: int,
        mock_esi: MagicMock,
        *,
        contacts: Iterable[EsiContact] = None,
        labels: Iterable[EsiContactLabel] = None,
    ) -> "EsiCharacterContactsStub":
        """Create new obj for tests."""
        obj = cls(character_id, EsiContactsContainer())
        if labels:
            obj._setup_labels(labels)
        if contacts:
            obj._setup_contacts(contacts)
        if mock_esi:
            obj._setup_esi_mock(mock_esi)
        return obj

    def _fetch_contacts(self, token: Token):
        self._assert_correct_character(token.character_id)
        return self._contacts.contacts()

    def _fetch_labels(self, token: Token):
        self._assert_correct_character(token.character_id)
        return self._contacts.labels()

    def _add_contacts(self, token: Token, contacts: Iterable[EsiContact]):
        self._assert_correct_character(token.character_id)
        label_ids = set()
        for c in contacts:
            label_ids |= c.label_ids
        self._check_label_ids_valid(label_ids)
        for c in contacts:
            self._contacts.add_contact(c)

    def _update_contacts(self, token: Token, contacts: Iterable[EsiContact]):
        self._assert_correct_character(token.character_id)
        label_ids = set()
        for c in contacts:
            label_ids |= c.label_ids
        for c in contacts:
            self._contacts.update_contact(c)

    def _delete_contacts(self, token: Token, contacts: Iterable[EsiContact]):
        self._assert_correct_character(token.character_id)
        for c in contacts:
            self._contacts.remove_contact(c)

    def _assert_correct_character(self, character_id: int):
        if character_id != self._character_id:
            raise ValueError(f"Unknown character ID: {character_id}")

    def _check_label_ids_valid(self, label_ids: Iterable[int]):
        for label_id in label_ids:
            self._contacts.label_by_id(label_id)


class TestCaseWithClearCache(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cache.clear()


def extract(qs: QuerySet, field: str) -> Set[int]:
    """Return the extracted fields from the items of a query set."""
    return set(qs.values_list(field, flat=True))



from unittest.mock import MagicMock, patch

from celery.exceptions import Retry

from django.core.cache import cache
from django.test import TestCase, override_settings
from django.utils.timezone import now
from esi.exceptions import ESIBucketLimitException
from esi.rate_limiting import ESIRateLimitBucket

from app_utils.testing import NoSocketsTestCase

from standingssync import tasks
from standingssync.models import SyncManager
from standingssync.tests.factories import (
    EveContactFactory,
    EveWarEmptyFactory,
    EveWarFactory,
    SyncedCharacterFactory,
    SyncManagerFactory,
    UserMainManagerFactory,
)

MANAGERS_PATH = "standingssync.managers"
MODELS_PATH = "standingssync.models"
TASKS_PATH = "standingssync.tasks"


@patch(TASKS_PATH + ".run_manager_sync")
@patch(TASKS_PATH + ".sync_all_wars")
class TestRunRegularSync(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = UserMainManagerFactory()

    def setUp(self):
        cache.clear()

    def test_should_not_sync_wars_if_disabled(
        self, mock_update_all_wars, mock_run_manager_sync
    ):
        # when
        with (
            patch(TASKS_PATH + ".is_esi_online", lambda: True),
            patch(TASKS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", False),
        ):
            tasks.run_regular_sync()
        # then
        self.assertFalse(mock_update_all_wars.apply_async.called)

    def test_should_start_all_tasks(self, mock_update_all_wars, mock_run_manager_sync):
        # given
        sync_manager = SyncManagerFactory(user=self.user)
        # when
        with (
            patch(TASKS_PATH + ".is_esi_online", lambda: True),
            patch(TASKS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True),
        ):
            tasks.run_regular_sync()
        # then
        self.assertTrue(mock_update_all_wars.apply_async.called)
        _, kwargs = mock_run_manager_sync.apply_async.call_args
        self.assertListEqual(kwargs["args"], [sync_manager.pk])

    def test_abort_when_esi_if_offline(
        self, mock_update_all_wars, mock_run_manager_sync
    ):
        # when
        with patch(TASKS_PATH + ".is_esi_online", lambda: False):
            tasks.run_regular_sync()
        # then
        self.assertFalse(mock_update_all_wars.apply_async.called)
        self.assertFalse(mock_run_manager_sync.apply_async.called)


@patch(TASKS_PATH + ".SyncedCharacter.run_sync")
class TestCharacterSync(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.synced_character = SyncedCharacterFactory()

    def setUp(self):
        cache.clear()

    def test_should_call_update(self, mock_update):
        # given
        mock_update.return_value = True
        # when
        tasks.run_character_sync(self.synced_character.pk)
        # then
        self.assertTrue(mock_update.called)

    def test_should_raise_exception(self, mock_update):
        # given
        mock_update.side_effect = RuntimeError
        # when
        with self.assertRaises(RuntimeError):
            tasks.run_character_sync(self.synced_character.pk)


@patch(TASKS_PATH + ".run_character_sync")
class TestManagerSync(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user_manager = UserMainManagerFactory()

    def setUp(self):
        cache.clear()

    # run for non existing sync manager
    def test_run_sync_wrong_pk(self, mock_run_character_sync):
        with self.assertRaises(SyncManager.DoesNotExist):
            tasks.run_manager_sync(99999)

    @patch(MODELS_PATH + ".SyncManager.run_sync")
    def test_should_abort_when_unexpected_exception_occurs(
        self, mock_update_from_esi, mock_run_character_sync
    ):
        # given
        mock_update_from_esi.side_effect = RuntimeError
        sm = SyncManagerFactory(user=self.user_manager)

        # when/then
        with self.assertRaises(RuntimeError):
            tasks.run_manager_sync(sm.pk)

    @patch(MODELS_PATH + ".SyncManager.run_sync")
    def test_should_normally_run_character_sync(
        self, mock_update_from_esi, mock_run_character_sync
    ):
        # given
        mock_update_from_esi.return_value = "abc"
        sm = SyncManagerFactory(user=self.user_manager)
        sc = SyncedCharacterFactory(manager=sm)

        # when
        tasks.run_manager_sync(sm.pk)

        # then
        sm.refresh_from_db()
        _, kwargs = mock_run_character_sync.apply_async.call_args
        self.assertEqual(kwargs["kwargs"]["pk"], sc.pk)

    @patch(MODELS_PATH + ".SyncManager.run_sync")
    def test_should_abort_when_too_many_contacts(
        self, mock_update_from_esi, mock_run_character_sync
    ):
        # given
        mock_update_from_esi.return_value = "abc"
        sm = SyncManagerFactory(user=self.user_manager)

        for _ in range(1025):
            EveContactFactory(manager=sm)

        # when/then
        with self.assertRaises(RuntimeError):
            tasks.run_manager_sync(sm.pk)


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(TASKS_PATH + ".sync_stale_war")
@patch(TASKS_PATH + ".EveWar.objects.sync_known_wars")
class TestSyncAllWars(TestCase):
    def setUp(self):
        cache.clear()

    def test_should_start_sync_wars_task_when_wars_to_sync(
        self, mock_calc_relevant_war_ids: MagicMock, mock_sync_stale_war: MagicMock
    ):
        # given
        EveWarEmptyFactory()
        # when
        tasks.sync_all_wars.delay()
        # then
        self.assertEqual(mock_sync_stale_war.apply_async.call_count, 1)

    def test_should_not_start_sync_wars_tasks_when_no_wars_to_sync(
        self, mock_calc_relevant_war_ids: MagicMock, mock_sync_stale_war: MagicMock
    ):
        # given
        EveWarFactory(finished=now())
        # when
        tasks.sync_all_wars.delay()
        # then
        self.assertEqual(mock_sync_stale_war.apply_async.call_count, 0)


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(TASKS_PATH + ".EveWar.update_from_esi")
class TestSyncWars(NoSocketsTestCase):
    def setUp(self):
        cache.clear()

    def test_should_update_war(self, mock_update_from_esi: MagicMock):
        # given
        war = EveWarEmptyFactory()

        def update():
            war.finished = now()
            war.save()

        mock_update_from_esi.side_effect = update
        # when
        tasks.sync_stale_war.delay()
        # then
        self.assertEqual(mock_update_from_esi.call_count, 1)

    def test_should_exit_when_no_wars_to_update(self, mock_update_from_esi: MagicMock):
        # given
        EveWarFactory(finished=now())
        # when
        tasks.sync_stale_war.delay()
        # then
        self.assertEqual(mock_update_from_esi.call_count, 0)

    def test_should_retry_on_rate_limit_exhausted(self, mock_update_from_esi):
        # given
        bucket = ESIRateLimitBucket("dummy", 10, 3600)
        ex = ESIBucketLimitException(bucket)
        mock_update_from_esi.side_effect = ex
        EveWarEmptyFactory()
        # when
        with self.assertRaises(Retry):
            tasks.sync_stale_war.delay()



from unittest.mock import Mock, patch

from django.contrib.auth.models import User
from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory, TestCase
from django.urls import reverse
from esi.models import Token
from eveuniverse.tests.testdata.factories_2 import EveEntityCharacterFactory

from allianceauth.eveonline.models import EveCharacter
from app_utils.testdata_factories import (
    EveAllianceInfoFactory,
    EveCharacterFactory,
    UserMainFactory,
)
from app_utils.testing import NoSocketsTestCase, add_character_to_user

from standingssync import views
from standingssync.models import SyncedCharacter, SyncManager
from standingssync.tests.factories import (
    EveContactFactory,
    SyncedCharacterFactory,
    SyncManagerFactory,
    UserMainDefaultFactory,
    UserMainManagerFactory,
)

MODULE_PATH = "standingssync.views"


class TestMainScreen(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.factory = RequestFactory()

        cls.user_manager = UserMainManagerFactory()
        cls.sync_manager = SyncManagerFactory(user=cls.user_manager)
        cls.user_normal = UserMainDefaultFactory(
            main_character__alliance_id=cls.sync_manager.alliance.alliance_id
        )
        cls.sync_char = SyncedCharacterFactory(
            manager=cls.sync_manager, user=cls.user_manager
        )

        cls.user_no_permission = UserMainFactory()

    def test_should_redirect_to_main_page(self):
        # given
        request = self.factory.get(reverse("standingssync:characters"))
        request.user = self.user_normal
        # when
        response = views.index(request)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("standingssync:characters"))

    def test_user_with_permission_can_open_app(self):
        request = self.factory.get(reverse("standingssync:characters"))
        request.user = self.user_normal
        response = views.characters(request)
        self.assertEqual(response.status_code, 200)

    def test_user_wo_permission_can_not_open_app(self):
        request = self.factory.get(reverse("standingssync:characters"))
        request.user = self.user_no_permission
        response = views.characters(request)
        self.assertEqual(response.status_code, 302)

    @patch(MODULE_PATH + ".messages")
    def test_user_can_remove_sync_char(self, mock_messages):
        request = self.factory.get(
            reverse("standingssync:remove_character", args=(self.sync_char.pk,))
        )
        request.user = self.user_normal
        response = views.remove_character(request, self.sync_char.pk)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(mock_messages.success.called)
        self.assertFalse(SyncedCharacter.objects.filter(pk=self.sync_char.pk).exists())


@patch(MODULE_PATH + ".tasks.run_character_sync")
@patch(MODULE_PATH + ".messages")
class TestAddSyncChar(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()

        alliance = EveAllianceInfoFactory()
        character = EveCharacterFactory(corporation__alliance=alliance)
        cls.user_manager = UserMainManagerFactory(main_character__character=character)
        cls.sync_manager = SyncManagerFactory(user=cls.user_manager)

        cls.character_normal = EveCharacterFactory(corporation__alliance=alliance)
        cls.user_normal = UserMainDefaultFactory(
            main_character__character=cls.character_normal
        )

    def make_request(self, user: User, character: EveCharacter):
        token: Token = user.token_set.get(character_id=character.character_id)
        token.character_id = character.character_id
        request = self.factory.get(reverse("standingssync:add_character"))
        request.user = user
        request.token = token  # type: ignore
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        orig_view = views.add_character.__wrapped__.__wrapped__.__wrapped__
        return orig_view(request, token)

    def test_user_can_add_blue_alt(self, mock_messages, mock_run_character_sync):
        # given
        alt_character = EveCharacterFactory()
        add_character_to_user(self.user_normal, alt_character)
        EveContactFactory(
            manager=self.sync_manager,
            eve_entity=EveEntityCharacterFactory(id=alt_character.character_id),
            standing=10,
        )

        # when
        response = self.make_request(self.user_normal, alt_character)

        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("standingssync:characters"))
        self.assertTrue(mock_messages.success.called)
        self.assertTrue(mock_run_character_sync.delay.called)
        self.assertTrue(
            SyncedCharacter.objects.filter(manager=self.sync_manager)
            .filter(character_ownership__character=alt_character)
            .exists()
        )

    def test_users_can_not_add_alliance_members(
        self, mock_messages, mock_run_character_sync
    ):
        # when
        with patch(MODULE_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0.1):
            response = self.make_request(self.user_normal, self.character_normal)

        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("standingssync:characters"))
        self.assertTrue(mock_messages.warning.called)
        self.assertFalse(mock_run_character_sync.delay.called)

    @patch(MODULE_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0)
    def test_user_can_add_neutral_alt(self, mock_messages, mock_run_character_sync):
        # given
        alt_character = EveCharacterFactory()
        add_character_to_user(self.user_normal, alt_character)

        # when
        with patch(MODULE_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0.0):
            response = self.make_request(self.user_normal, alt_character)

        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("standingssync:characters"))
        self.assertTrue(mock_messages.success.called)
        self.assertTrue(mock_run_character_sync.delay.called)
        self.assertTrue(
            SyncedCharacter.objects.filter(manager=self.sync_manager)
            .filter(character_ownership__character=alt_character)
            .exists()
        )

    def test_user_can_not_add_non_blue_alt(
        self, mock_messages, mock_run_character_sync
    ):
        # given
        alt_character = EveCharacterFactory()
        add_character_to_user(self.user_normal, alt_character)

        # when
        with patch(MODULE_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0.1):
            response = self.make_request(self.user_normal, alt_character)

        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("standingssync:characters"))
        self.assertTrue(mock_messages.warning.called)
        self.assertFalse(mock_run_character_sync.delay.called)
        self.assertFalse(
            SyncedCharacter.objects.filter(manager=self.sync_manager)
            .filter(character_ownership__character=alt_character)
            .exists()
        )


@patch(MODULE_PATH + ".tasks")
@patch(MODULE_PATH + ".messages")
class TestAddAllianceManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()

    def make_request(self, user: User):
        character: EveCharacter = user.profile.main_character
        token: Token = user.token_set.get(character_id=character.character_id)
        token.character_id = character.character_id
        request = self.factory.get(reverse("standingssync:add_alliance_manager"))
        request.user = user
        request.token = token  # type: ignore
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        orig_view = views.add_alliance_manager.__wrapped__.__wrapped__.__wrapped__
        return orig_view(request, token)

    def test_user_with_permission_can_add_alliance_manager(
        self, mock_messages, mock_tasks
    ):
        # given
        user = UserMainManagerFactory()

        # when
        response = self.make_request(user)

        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("standingssync:index"))
        self.assertTrue(mock_messages.success.called)
        self.assertTrue(mock_tasks.run_manager_sync.delay.called)
        self.assertTrue(SyncManager.objects.exists())

    # def test_user_wo_permission_can_not_add_alliance_manager(
    #     self, mock_messages, mock_tasks
    # ):
    #     # given
    #     user = UserMainSyncerFactory()

    #     # when
    #     response = self.make_request(user)

    #     # then
    #     self.assertEqual(response.status_code, 302)
    #     self.assertEqual(response.url, reverse("standingssync:index"))
    #     self.assertFalse(mock_messages.success.called)
    #     self.assertFalse(mock_tasks.run_manager_sync.delay.called)
    #     self.assertFalse(SyncManager.objects.exists())

    def test_should_sync_wars_when_adding_alliance_char_and_feature_enabled(
        self, mock_messages, mock_tasks
    ):
        # given
        user = UserMainManagerFactory()

        # when
        with patch(MODULE_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True):
            self.make_request(user)

        # then
        self.assertTrue(mock_tasks.sync_all_wars.delay.called)

    def test_should_not_sync_wars_when_adding_alliance_char_and_feature_disabled(
        self, mock_messages, mock_tasks
    ):
        # given
        user = UserMainManagerFactory()

        # when
        with patch(MODULE_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", False):
            self.make_request(user)

        # then
        self.assertFalse(mock_tasks.sync_all_wars.delay.called)

    def test_character_for_manager_must_be_alliance_member(
        self, mock_messages, mock_tasks
    ):
        # given
        character = EveCharacterFactory(alliance_id=None, alliance_name="")
        user = UserMainManagerFactory(main_character__character=character)

        # when
        response = self.make_request(user)

        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("standingssync:index"))
        self.assertTrue(mock_messages.warning.called)
        self.assertFalse(mock_tasks.run_manager_sync.delay.called)
        self.assertFalse(SyncManager.objects.exists())



from django.template import Context, Template
from django.test import TestCase
from eveuniverse.models import EveEntity
from eveuniverse.tests.testdata.factories_2 import (
    EveEntityAllianceFactory,
    EveEntityCharacterFactory,
    EveEntityCorporationFactory,
)


class TestTemplateTags(TestCase):
    def test_should_render_alliance_war_participant(self):
        # given
        template = Template("""
            {% load standingssync %}
            {% war_participant obj %}
            """)
        params = [
            (EveEntityAllianceFactory(), "alliance"),
            (EveEntityCorporationFactory(), "corporation"),
            (EveEntityCharacterFactory(), ""),
        ]
        for obj, expected in params:
            with self.subTest(category=obj.category):
                context = Context({"obj": obj})
                # when
                result = template.render(context)
                # then
                self.assertIn(str(obj.id), result)
                self.assertIn(expected, result)

    def test_should_display_no_data_for_empty_obj(self):
        # given
        template = Template("""
            {% load standingssync %}
            {% war_participant obj %}
            """)
        obj = EveEntity.objects.create(id=1)
        context = Context({"obj": obj})
        # when
        result = template.render(context)
        # then
        self.assertIn("[NO DATA]", result)

    def test_should_display_no_data_for_invalid_obj(self):
        # given
        template = Template("""
            {% load standingssync %}
            {% war_participant obj %}
            """)
        context = Context({"obj": "abc"})
        # when
        result = template.render(context)
        # then
        self.assertIn("[NO DATA]", result)



"""Model test factories."""

import datetime as dt
import urllib.parse
from typing import Generic, TypeVar

import factory
import factory.fuzzy

from django.utils.timezone import now
from eveuniverse.tests.testdata.factories_2 import (
    EveEntityAllianceFactory,
    EveEntityCharacterFactory,
)

from app_utils.testdata_factories import (
    EveAllianceInfoFactory,
    EveCharacterFactory,
    UserMainFactory,
)
from app_utils.testing import add_character_to_user

from standingssync.core.esi_contacts import EsiContact, EsiContactLabel
from standingssync.models import EveContact, EveWar, SyncedCharacter, SyncManager

T = TypeVar("T")
_BASE_URL = "https://esi.evetech.net/"


def make_esi_url(path: str) -> str:
    if path.startswith("/"):
        raise ValueError("path can not start with a slash")
    if path.endswith("/"):
        raise ValueError("path can not end with a slash")

    url = urllib.parse.urljoin(_BASE_URL, path)
    return url


class BaseMetaFactory(Generic[T], factory.base.FactoryMetaClass):
    def __call__(cls, *args, **kwargs) -> T:
        return super().__call__(*args, **kwargs)


class EveWarFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[EveWar]
):
    class Meta:
        model = EveWar

    class Params:
        is_finished = factory.Trait(finished=now())

    id = factory.Sequence(lambda n: 1 + n)
    aggressor = factory.SubFactory(EveEntityAllianceFactory)
    declared = factory.fuzzy.FuzzyDateTime(
        now() - dt.timedelta(days=3), end_dt=now() - dt.timedelta(days=2)
    )
    defender = factory.SubFactory(EveEntityAllianceFactory)
    is_mutual = False
    is_open_for_allies = True
    started = factory.LazyAttribute(lambda obj: obj.declared + dt.timedelta(hours=24))
    retracted = None

    @factory.lazy_attribute
    def finished(self):
        return self.retracted + dt.timedelta(hours=24) if self.retracted else None

    @factory.post_generation
    def allies(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for ally in extracted:
                self.allies.add(ally)  # type: ignore


class EveWarEmptyFactory(EveWarFactory):
    aggressor = None
    declared = None
    defender = None
    is_mutual = None
    is_open_for_allies = None
    started = None
    retracted = None


class UserMainManagerFactory(UserMainFactory):
    main_character__scopes = ["esi-alliances.read_contacts.v1"]
    permissions__ = ["standingssync.add_syncmanager"]


class UserMainDefaultFactory(UserMainFactory):
    main_character__scopes = [
        "esi-characters.read_contacts.v1",
        "esi-characters.write_contacts.v1",
    ]
    permissions__ = ["standingssync.add_syncedcharacter"]


class SyncManagerFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[SyncManager]
):
    class Meta:
        model = SyncManager

    class Params:
        user = factory.SubFactory(UserMainManagerFactory)

    version_hash = factory.fuzzy.FuzzyText(length=32)

    @factory.lazy_attribute
    def alliance(self):
        return EveAllianceInfoFactory(
            alliance_id=self.user.profile.main_character.alliance_id  # type: ignore
        )

    @factory.lazy_attribute
    def character_ownership(self):
        return self.user.profile.main_character.character_ownership  # type: ignore

    @factory.post_generation
    def create_eve_entities(self, create, extracted, **kwargs):
        if not create:
            return
        EveEntityAllianceFactory(
            id=self.alliance.alliance_id, name=self.alliance.alliance_name  # type: ignore
        )


class SyncedCharacterFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[SyncedCharacter]
):
    class Meta:
        model = SyncedCharacter

    class Params:
        user = None

    manager = factory.SubFactory(SyncManagerFactory)

    @factory.lazy_attribute
    def character_ownership(self):
        if not self.user:
            character = EveCharacterFactory(corporation__alliance=self.manager.alliance)
            user = UserMainDefaultFactory(main_character__character=character)
            return user.profile.main_character.character_ownership  # type: ignore

        return self.user.profile.main_character.character_ownership  # type: ignore

    @factory.post_generation
    def create_alt(self, create, extracted, **kwargs):
        if not create or extracted is not True:
            return

        alt = EveCharacterFactory()
        add_character_to_user(self.character_ownership.user, alt)


class EveContactFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[EveContact]
):
    class Meta:
        model = EveContact

    manager = factory.SubFactory(SyncManagerFactory)
    eve_entity = factory.SubFactory(EveEntityCharacterFactory)
    standing = factory.LazyAttribute(lambda o: -10 if o.is_war_target else 5)
    is_war_target = False


class EsiContactDictFactory(factory.base.DictFactory, metaclass=BaseMetaFactory[dict]):
    contact_id = factory.fuzzy.FuzzyInteger(90_000, 99_999)
    contact_type = factory.fuzzy.FuzzyChoice(["character", "corporation", "alliance"])
    standing = factory.fuzzy.FuzzyFloat(-10.0, 10.0)


class EsiLabelDictFactory(factory.base.DictFactory, metaclass=BaseMetaFactory[dict]):
    label_id = factory.fuzzy.FuzzyInteger(1, 9_999)
    label_name = factory.Faker("word")


class EsiContactFactory(factory.base.Factory, metaclass=BaseMetaFactory[EsiContact]):
    class Meta:
        model = EsiContact

    contact_id = factory.fuzzy.FuzzyInteger(90_000, 99_999)
    contact_type = factory.fuzzy.FuzzyChoice(list(EsiContact.Category))
    standing = factory.fuzzy.FuzzyFloat(-10.0, 10.0)


class EsiContactAllianceFactory(
    factory.base.Factory, metaclass=BaseMetaFactory[EsiContact]
):
    class Meta:
        model = EsiContact

    contact_id = factory.Sequence(lambda n: 99_909_001 + n)
    contact_type = EsiContact.Category.ALLIANCE
    standing = factory.fuzzy.FuzzyFloat(-10.0, 10.0)


class EsiContactCharacterFactory(
    factory.base.Factory, metaclass=BaseMetaFactory[EsiContact]
):
    class Meta:
        model = EsiContact

    contact_id = factory.Sequence(lambda n: 90_909_001 + n)
    contact_type = EsiContact.Category.CHARACTER
    standing = factory.fuzzy.FuzzyFloat(-10.0, 10.0)


class EsiContactCorporationFactory(
    factory.base.Factory, metaclass=BaseMetaFactory[EsiContact]
):
    class Meta:
        model = EsiContact

    contact_id = factory.Sequence(lambda n: 97_909_001 + n)
    contact_type = EsiContact.Category.CORPORATION
    standing = factory.fuzzy.FuzzyFloat(-10.0, 10.0)


class EsiContactFactionFactory(
    factory.base.Factory, metaclass=BaseMetaFactory[EsiContact]
):
    class Meta:
        model = EsiContact

    contact_id = factory.Sequence(lambda n: 509_901 + n)
    contact_type = EsiContact.Category.FACTION
    standing = factory.fuzzy.FuzzyFloat(-10.0, 10.0)


class EsiContactLabelFactory(
    factory.base.Factory, metaclass=BaseMetaFactory[EsiContactLabel]
):
    class Meta:
        model = EsiContactLabel

    id = factory.fuzzy.FuzzyInteger(1, 9_999)
    name = factory.Faker("word")



"""Models for standingssync."""

import datetime as dt
from typing import Optional, Set

from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.db import models, transaction
from django.utils.timezone import now
from esi.errors import TokenExpiredError, TokenInvalidError
from esi.models import Token
from eveuniverse.models import EveEntity

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveAllianceInfo, EveCharacter
from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger

from standingssync.app_settings import (
    STANDINGSSYNC_ADD_WAR_TARGETS,
    STANDINGSSYNC_CHAR_MIN_STANDING,
    STANDINGSSYNC_REPLACE_CONTACTS,
    STANDINGSSYNC_STORE_ESI_CONTACTS_ENABLED,
    STANDINGSSYNC_SYNC_TIMEOUT,
)
from standingssync.core import esi_api
from standingssync.core.esi_contacts import EsiContact, EsiContactsContainer
from standingssync.helpers import store_json
from standingssync.managers import EveContactManager, EveWarManager, SyncManagerManager

logger = get_extension_logger(__name__)


class _SyncBaseModel(models.Model):
    """Common fields and logic for sync models."""

    last_sync_at = models.DateTimeField(
        null=True,
        default=None,
        help_text="When the last successful sync was completed.",
    )

    class Meta:
        abstract = True

    @property
    def is_sync_fresh(self) -> bool:
        """Return True, when sync is not stale, else False."""
        if not self.last_sync_at:
            return False

        deadline = now() - dt.timedelta(minutes=STANDINGSSYNC_SYNC_TIMEOUT)
        return self.last_sync_at > deadline

    def record_successful_sync(self):
        """Record date & time of a successful sync."""
        self.last_sync_at = now()
        self.save()


class SyncManager(_SyncBaseModel):
    """An object for managing syncing of contacts for an alliance."""

    alliance = models.OneToOneField(
        EveAllianceInfo,
        on_delete=models.CASCADE,
        primary_key=True,
        related_name="+",
        help_text="Alliance to sync contacts for",
    )
    character_ownership = models.OneToOneField(
        CharacterOwnership,
        on_delete=models.SET_NULL,
        null=True,
        default=None,
        help_text="alliance contacts are fetched from this character",
    )
    version_hash = models.CharField(
        max_length=32,
        default="",
        help_text="hash over all contacts to identify when it has changed",
    )
    compress_contacts = models.BooleanField(
        default=False,
        help_text=(
            "Remove contacts which are unnecessary for calculating "
            "their effective standing."
        ),
    )
    exclude_character_contacts = models.BooleanField(
        default=False,
        help_text="Exclude all character contacts",
    )
    exclude_negative_standings = models.BooleanField(
        default=False,
        help_text="Exclude all contacts with negative standing",
    )
    exclude_positive_standings = models.BooleanField(
        default=False,
        help_text="Exclude all contacts with positive standing",
    )
    unmask_war_targets = models.BooleanField(
        default=False,
        help_text=(
            "Remove contacts which would mask the effective standing of "
            "war target contacts"
        ),
    )

    objects = SyncManagerManager()

    def __str__(self):
        return str(self.alliance)

    @property
    def character(self) -> EveCharacter:
        """Return character linked to this manager or raises exception of none exists."""
        if not self.character_ownership:
            raise ValueError("No character ownership")
        return self.character_ownership.character

    def contacts_for_sync(self, synced_character: "SyncedCharacter") -> models.QuerySet:
        """Relevant contacts for sync, which excludes the sync character."""
        return self.contacts.exclude(eve_entity_id=synced_character.character_id)

    def effective_standing_with_character(self, character: EveCharacter) -> float:
        """Effective standing of the alliance with a character."""
        try:
            return self.contacts.get(eve_entity_id=character.character_id).standing
        except EveContact.DoesNotExist:
            pass

        try:
            return self.contacts.get(eve_entity_id=character.corporation_id).standing
        except EveContact.DoesNotExist:
            pass

        if character.alliance_id:
            try:
                return self.contacts.get(eve_entity_id=character.alliance_id).standing
            except EveContact.DoesNotExist:
                pass

        return 0.0

    def synced_characters_for_user(
        self, user: User
    ) -> models.QuerySet["SyncedCharacter"]:
        """Synced characters of the given user."""
        return self.synced_characters.filter(character_ownership__user=user)

    def run_sync(self, force_update: bool = False) -> bool:
        """Run sync for this manager.

        Args:
            - force_update: when true will always update contacts in database

        Returns True when contacts where updated, else False.
        """
        if self.character_ownership is None:
            raise RuntimeError(f"{self}: Can not sync. No character configured.")

        if not self.character_ownership.user.has_perm("standingssync.add_syncmanager"):
            raise RuntimeError(
                f"{self}: Can not sync. Character does not have sufficient permission."
            )

        token = self.fetch_token()
        if not token:
            raise RuntimeError(f"{self}: Can not sync. No valid token found.")

        esi_contacts = esi_api.fetch_alliance_contacts(self.alliance.alliance_id, token)
        if self.exclude_positive_standings:
            esi_contacts = {x for x in esi_contacts if x.standing <= 0}

        if self.exclude_negative_standings:
            esi_contacts = {x for x in esi_contacts if x.standing >= 0}

        if self.exclude_character_contacts:
            esi_contacts = {
                x
                for x in esi_contacts
                if x.contact_type != EsiContact.Category.CHARACTER
            }

        contacts = EsiContactsContainer.from_esi_contacts(esi_contacts)
        war_target_ids = self._add_war_targets(contacts)

        unmask_war_targets = STANDINGSSYNC_ADD_WAR_TARGETS and self.unmask_war_targets
        if self.compress_contacts or unmask_war_targets:
            removed_count = contacts.prune(
                unmask_war_targets=unmask_war_targets,
                compress_contacts=self.compress_contacts,
            )
            logger.info("%s: Pruned contacts. Removed %d", self, removed_count)

        new_version_hash = contacts.version_hash()

        if force_update or new_version_hash != self.version_hash:
            self._save_new_contacts(contacts, war_target_ids, new_version_hash)
            EveEntity.objects.bulk_resolve_ids(contacts.contact_ids())
            was_updated = True
        else:
            logger.info("%s: Alliance contacts are unchanged.", self)
            was_updated = False

        self.record_successful_sync()
        return was_updated

    def fetch_token(self) -> Optional[Token]:
        """Fetch valid token with required scopes."""
        if not self.character_ownership:
            return None

        token = (
            Token.objects.filter(
                user=self.character_ownership.user,
                character_id=self.character.character_id,
            )
            .require_scopes(self.get_esi_scopes())
            .require_valid()
            .first()
        )
        return token

    def _add_war_targets(self, contacts: EsiContactsContainer) -> Set[int]:
        """Add war targets to contacts (if enabled).

        Returns contact IDs of war targets.
        """
        if not STANDINGSSYNC_ADD_WAR_TARGETS:
            return set()

        war_targets = EveWar.objects.alliance_war_targets(self.alliance)
        war_target_ids = set()
        for war_target in war_targets:
            try:
                wt = EsiContact.from_eve_entity(
                    war_target, standing=-10.0, is_war_target=True
                )
                contacts.add_contact(wt)
            except ValueError:  # eve_entity has no category
                logger.warning("Skipping unresolved war target: %s", war_target)
            else:
                war_target_ids.add(war_target.id)

        return war_target_ids

    # TODO: Replace delete & create with update & delete
    @transaction.atomic()
    def _save_new_contacts(
        self,
        current_contacts: EsiContactsContainer,
        war_target_ids: Set[int],
        new_version_hash: str,
    ):
        self.contacts.all().delete()
        contacts = [
            EveContact(
                manager=self,
                eve_entity=EveEntity.objects.get_or_create(id=contact.contact_id)[0],
                standing=contact.standing,
                is_war_target=contact.contact_id in war_target_ids,
            )
            for contact in current_contacts.contacts()
        ]
        EveContact.objects.bulk_create(contacts, batch_size=500)
        self.version_hash = new_version_hash
        self.save()
        logger.info("%s: Stored alliance update with %d contacts", self, len(contacts))

    @classmethod
    def get_esi_scopes(cls) -> list:
        """Return ESI scopes needed for this model."""
        return ["esi-alliances.read_contacts.v1"]


class SyncedCharacter(_SyncBaseModel):
    """A character that has his personal contacts synced with an alliance"""

    character_ownership = models.OneToOneField(
        CharacterOwnership,
        on_delete=models.CASCADE,
        primary_key=True,
        help_text="Character to sync",
    )
    has_war_targets_label = models.BooleanField(
        default=None,
        null=True,
        help_text="Whether this character has the war target label.",
    )
    manager = models.ForeignKey(
        SyncManager, on_delete=models.CASCADE, related_name="synced_characters"
    )

    def __str__(self):
        try:
            character_name = self.character.character_name
        except ObjectDoesNotExist:
            character_name = "?"
        return f"{character_name} - {self.manager}"

    @property
    def character(self) -> EveCharacter:
        """Return EveCharacter used to sync."""
        return self.character_ownership.character

    @property
    def character_id(self) -> Optional[int]:
        """Return character ID of the EveCharacter used to sync."""
        return self.character.character_id if self.character else None

    def run_sync(self) -> Optional[bool]:
        """Sync in-game contacts for given character with alliance contacts
        and/or war targets.

        Will delete this sync character if necessary,
        e.g. if token is no longer valid or character is no longer blue

        Returns:
        - False when the sync character was deleted
        - None when no update was needed
        - True when update was done successfully
        """
        if not self._has_sync_permissions():
            return False

        if not self._has_standing_with_alliance():
            return False

        token = self.fetch_token()
        if not token:
            logger.error("%s: Can not sync. No valid token found.", self)
            return False

        if not self.manager.contacts_for_sync(self).exists():
            logger.info("%s: No contacts to sync", self)
            return None

        current_contacts = self._fetch_current_contacts(token)
        self._update_wt_label_info(current_contacts)

        new_contacts = self._identify_new_contacts(current_contacts)

        self._update_contacts_on_esi(token, current_contacts, new_contacts)

        if STANDINGSSYNC_STORE_ESI_CONTACTS_ENABLED:
            store_json(new_contacts.to_dict(), "new_contacts")

        self.record_successful_sync()
        return True

    def _identify_new_contacts(self, current_contacts):
        if STANDINGSSYNC_REPLACE_CONTACTS:
            new_contacts = EsiContactsContainer.from_esi_contacts(
                labels=current_contacts.labels()
            )
            new_contacts.add_eve_contacts(
                self.manager.contacts_for_sync(self).filter(is_war_target=False)
            )
        else:
            new_contacts = current_contacts.clone()

        if STANDINGSSYNC_ADD_WAR_TARGETS:
            # remove old war targets
            new_contacts.remove_war_targets()
            # add current war targets
            wt_label_id = current_contacts.war_target_label_id()
            new_contacts.add_eve_contacts(
                self.manager.contacts_for_sync(self).filter(is_war_target=True),
                label_ids=[wt_label_id] if wt_label_id else None,
            )

        return new_contacts

    def _update_contacts_on_esi(self, token, current_contacts, new_contacts):
        added, removed, changed = current_contacts.contacts_difference(new_contacts)
        if removed:
            esi_api.delete_character_contacts(token, removed)
            logger.info("%s: Deleted %d added contacts", self, len(removed))

        if added:
            esi_api.add_character_contacts(token, added)
            logger.info("%s: Added %d missing contacts", self, len(added))

        if changed:
            esi_api.update_character_contacts(token, changed)
            logger.info("%s: Updated %d changed contacts", self, len(changed))

        if not added and not removed and not changed:
            logger.info("%s: Nothing updated. Contacts were already up-to-date.", self)
        else:
            logger.info("%s: Contacts update completed.", self)

    def _update_wt_label_info(self, current_contacts: EsiContactsContainer):
        """Update info about WT label if it has changed."""
        has_wt_label = current_contacts.war_target_label_id() is not None
        if has_wt_label != self.has_war_targets_label:
            self.has_war_targets_label = has_wt_label
            self.save()

    def _has_sync_permissions(self) -> bool:
        if not self.character_ownership.user.has_perm(
            "standingssync.add_syncedcharacter"
        ):
            logger.info(
                "%s: sync deactivated due to insufficient user permissions", self
            )
            self._deactivate_sync("you no longer have permission for this service")
            return False

        return True

    def fetch_token(self) -> Optional[Token]:
        """Fetch valid token with required scopes.

        Will deactivate this character if any severe issues are encountered.
        """
        try:
            token = self._valid_token()

        except TokenInvalidError:
            self._deactivate_sync("your token is no longer valid")
            logger.info("%s: sync deactivated due to invalid token", self)
            return None

        except TokenExpiredError:
            self._deactivate_sync("your token has expired")
            logger.info("%s: sync deactivated due to expired token", self)
            return None

        if token is None:
            self._deactivate_sync("you do not have a token anymore")
            logger.info("%s: can not find suitable token for synced char", self)
            return None

        return token

    def _valid_token(self) -> Optional[Token]:
        return (
            Token.objects.filter(
                user=self.character_ownership.user,
                character_id=self.character_ownership.character.character_id,
            )
            .require_scopes(self.get_esi_scopes())
            .require_valid()
            .first()
        )

    def _deactivate_sync(self, message: str):
        """Deactivate character and send a message to the user about the issue."""
        message = (
            "Standings Sync has been deactivated for your "
            f"character {self}, because {message}.\n"
            "Feel free to activate sync for your character again, "
            "once the issue has been resolved."
        )
        notify(
            self.character_ownership.user,
            f"Standings Sync deactivated for {self}",
            message,
        )
        self.delete()

    def _has_standing_with_alliance(self) -> bool:
        """Clarify if this character has standing with the alliance
        and therefore is allowed to sync contacts.

        Deactivate character if it has no standing.
        """
        character_eff_standing = self.manager.effective_standing_with_character(
            self.character_ownership.character
        )
        if character_eff_standing < STANDINGSSYNC_CHAR_MIN_STANDING:
            logger.info(
                "%s: sync deactivated because character is no longer considered blue. "
                f"It's standing is: {character_eff_standing}, "
                "while STANDINGSSYNC_CHAR_MIN_STANDING is: "
                f"{STANDINGSSYNC_CHAR_MIN_STANDING} ",
                self,
            )
            self._deactivate_sync(
                "your character is no longer blue with the alliance. "
                f"The standing value is: {character_eff_standing:.1f} "
            )
            return False

        return True

    def _fetch_current_contacts(self, token: Token) -> EsiContactsContainer:
        contacts = esi_api.fetch_character_contacts(token)
        labels = esi_api.fetch_character_contact_labels(token)
        current_contacts = EsiContactsContainer.from_esi_contacts(contacts, labels)

        if STANDINGSSYNC_STORE_ESI_CONTACTS_ENABLED:
            store_json(current_contacts.to_dict(), "current_contacts")

        logger.debug("%s: new version hash: %s", self, current_contacts.version_hash())
        return current_contacts

    def delete_all_contacts(self):
        """Delete all contacts of this character."""
        token = self._valid_token()
        if not token:
            logger.warning(
                "%s: Can not delete contacts, because no valid token found.", self
            )
            return

        contacts_clone = self._fetch_current_contacts(token)
        contacts = contacts_clone.contacts()
        esi_api.delete_character_contacts(token, contacts)
        logger.info("%s: Deleted all %d contacts", self, len(contacts))

    @staticmethod
    def get_esi_scopes() -> list:
        """Return required ESI scopes."""
        return ["esi-characters.read_contacts.v1", "esi-characters.write_contacts.v1"]


class EveContact(models.Model):
    """An Eve Online contact"""

    manager = models.ForeignKey(
        SyncManager, on_delete=models.CASCADE, related_name="contacts"
    )
    eve_entity = models.ForeignKey(
        EveEntity, on_delete=models.CASCADE, related_name="+"
    )
    standing = models.FloatField()
    is_war_target = models.BooleanField()

    objects = EveContactManager()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["manager", "eve_entity"], name="fk_eve_contact"
            )
        ]

    def __str__(self):
        return f"{self.eve_entity}"

    @property
    def contact_id(self) -> int:
        """Return contact ID."""
        return self.eve_entity_id


class EveWar(models.Model):
    """An EveOnline war"""

    class State(models.TextChoices):
        """A war state."""

        CONCLUDING = "concluding"  # active and about to finish normally
        FINISHED = "finished"  # finished war
        ONGOING = "ongoing"  # active and without finish date
        PENDING = "pending"  # declared, but not started yet
        RETRACTED = "retracted"  # activate and about to finish after retraction
        UNKNOWN = "unknown"  # state is unknown

        @classmethod
        def active_states(cls) -> Set["EveWar.State"]:
            """Return all states representing an active war."""
            return {cls.ONGOING, cls.CONCLUDING, cls.RETRACTED}

    id = models.PositiveIntegerField(primary_key=True)
    aggressor = models.ForeignKey(
        EveEntity, null=True, default=None, on_delete=models.SET_NULL, related_name="+"
    )
    allies = models.ManyToManyField(EveEntity, related_name="+")
    declared = models.DateTimeField(null=True, default=None)
    defender = models.ForeignKey(
        EveEntity, null=True, default=None, on_delete=models.SET_NULL, related_name="+"
    )
    finished = models.DateTimeField(null=True, default=None, db_index=True)
    is_mutual = models.BooleanField(null=True, default=None)
    is_open_for_allies = models.BooleanField(null=True, default=None)
    retracted = models.DateTimeField(null=True, default=None)
    started = models.DateTimeField(null=True, default=None, db_index=True)
    last_modified = models.DateTimeField(auto_now=True)

    objects = EveWarManager()

    def __str__(self) -> str:
        return f"{self.aggressor} vs. {self.defender}"

    def update_from_esi(self):
        """Update this war from ESI."""

        war_info = esi_api.fetch_war(war_id=self.id)
        entity_ids = set()
        aggressor = _get_or_create_eve_entity_from_participant(war_info["aggressor"])
        entity_ids.add(aggressor.id)
        defender = _get_or_create_eve_entity_from_participant(war_info["defender"])
        entity_ids.add(defender.id)
        self.aggressor = aggressor
        self.declared = war_info["declared"]
        self.defender = defender
        self.is_mutual = war_info["mutual"]
        self.is_open_for_allies = war_info["open_for_allies"]
        self.retracted = war_info.get("retracted")
        self.started = war_info.get("started")
        self.finished = war_info.get("finished")
        self.save()

        self.allies.clear()
        if war_info.get("allies"):
            for ally_info in war_info.get("allies", []):
                try:
                    ally = _get_or_create_eve_entity_from_participant(ally_info)
                except ValueError:
                    logger.warning("%s: Could not identify ally: %s", id, ally_info)
                    continue

                self.allies.add(ally)
                entity_ids.add(ally.id)

        EveEntity.objects.bulk_resolve_ids(entity_ids)


def _get_or_create_eve_entity_from_participant(participant: dict) -> EveEntity:
    """Get or create an EveEntity object from a war participant dict."""
    entity_id = participant.get("alliance_id") or participant.get("corporation_id")
    if not entity_id:
        raise ValueError(f"Invalid participant: {participant}")

    obj, _ = EveEntity.objects.get_or_create(id=entity_id)
    return obj



from django.apps import AppConfig

from standingssync import __version__


class SyncAltContactsConfig(AppConfig):
    name = "standingssync"
    label = "standingssync"
    verbose_name = f"Standings Sync v{__version__}"



# Generated by Django 4.0.7 on 2022-08-08 16:09

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    replaces = [
        ("standingssync", "0001_initial-new"),
        ("standingssync", "0002_add_war_targets"),
        ("standingssync", "0003_war_target_labels"),
        ("standingssync", "0004_remove_old_eve_entity"),
        ("standingssync", "0005_add_new_eve_entity"),
    ]

    initial = True

    dependencies = [
        ("eveonline", "0015_factions"),
        ("eveuniverse", "0007_evetype_description"),
        ("authentication", "0019_merge_20211026_0919"),
    ]

    operations = [
        migrations.CreateModel(
            name="SyncManager",
            fields=[
                ("version_hash", models.CharField(default="", max_length=32)),
                ("last_sync", models.DateTimeField(default=None, null=True)),
                (
                    "alliance",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        related_name="+",
                        serialize=False,
                        to="eveonline.eveallianceinfo",
                    ),
                ),
                (
                    "last_error",
                    models.IntegerField(
                        choices=[
                            (0, "No error"),
                            (1, "Invalid token"),
                            (2, "Expired token"),
                            (3, "Insufficient permissions"),
                            (4, "No character set for fetching alliance contacts"),
                            (5, "ESI API is currently unavailable"),
                            (99, "Unknown error"),
                        ],
                        default=0,
                    ),
                ),
                (
                    "character_ownership",
                    models.OneToOneField(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="authentication.characterownership",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="SyncedCharacter",
            fields=[
                ("version_hash", models.CharField(default="", max_length=32)),
                ("last_sync", models.DateTimeField(default=None, null=True)),
                (
                    "character_ownership",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        serialize=False,
                        to="authentication.characterownership",
                    ),
                ),
                (
                    "last_error",
                    models.IntegerField(
                        choices=[
                            (0, "No error"),
                            (1, "Invalid token"),
                            (2, "Expired token"),
                            (3, "Insufficient permissions"),
                            (5, "ESI API is currently unavailable"),
                            (99, "Unknown error"),
                        ],
                        default=0,
                    ),
                ),
                ("has_war_targets_label", models.BooleanField(default=None, null=True)),
                (
                    "manager",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="synced_characters",
                        to="standingssync.syncmanager",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="EveWar",
            fields=[
                ("id", models.PositiveIntegerField(primary_key=True, serialize=False)),
                ("declared", models.DateTimeField()),
                (
                    "finished",
                    models.DateTimeField(db_index=True, default=None, null=True),
                ),
                ("is_mutual", models.BooleanField()),
                ("is_open_for_allies", models.BooleanField()),
                ("retracted", models.DateTimeField(default=None, null=True)),
                (
                    "started",
                    models.DateTimeField(db_index=True, default=None, null=True),
                ),
                (
                    "aggressor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "allies",
                    models.ManyToManyField(
                        related_name="+", to="eveuniverse.eveentity"
                    ),
                ),
                (
                    "defender",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="EveContact",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("standing", models.FloatField()),
                ("is_war_target", models.BooleanField()),
                (
                    "eve_entity",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "manager",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="contacts",
                        to="standingssync.syncmanager",
                    ),
                ),
            ],
        ),
        migrations.AddConstraint(
            model_name="evecontact",
            constraint=models.UniqueConstraint(
                fields=("manager", "eve_entity"), name="fk_eve_contact"
            ),
        ),
    ]



"""Tasks for standingssync."""

import time

from celery import Task, shared_task

from django.db.models import QuerySet
from esi.decorators import rate_limit_retry_task
from eveuniverse.core.esitools import is_esi_online

from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce

from standingssync.app_settings import STANDINGSSYNC_ADD_WAR_TARGETS
from standingssync.constants import MAX_CHARACTER_CONTACTS
from standingssync.models import EveWar, SyncedCharacter, SyncManager

logger = get_extension_logger(__name__)


DEFAULT_TASK_PRIORITY = 6
ONCE_TIMEOUT = 1000  # once timeout determined by max rate limit reset + contingency
SYNC_WAR_DELAY = 0.6  # delay in seconds for fetching each war with contigency.


@shared_task(base=QueueOnce)
def run_regular_sync():
    """update all wars, managers and related characters if needed"""
    if not is_esi_online():
        logger.warning("ESI is not online. aborting")
        return

    if STANDINGSSYNC_ADD_WAR_TARGETS:
        sync_all_wars.apply_async(priority=DEFAULT_TASK_PRIORITY)

    for pk in SyncManager.objects.values_list("pk", flat=True):
        run_manager_sync.apply_async(args=[pk], priority=DEFAULT_TASK_PRIORITY)


@shared_task(base=QueueOnce, bind=True)
@rate_limit_retry_task
def run_manager_sync(_self, manager_pk: int, force_update: bool = False):
    """updates contacts for given manager and related characters

    Args:
    - manage_pk: primary key of sync manager to run sync for
    - force_update: will force update of manager even if not needed
    """
    sm = SyncManager.objects.get(pk=manager_pk)
    sm.run_sync(force_update)
    if sm.contacts.count() > MAX_CHARACTER_CONTACTS:
        raise RuntimeError(f"{sm}: Too many contacts")

    for character_pk in sm.synced_characters.values_list("pk", flat=True):
        run_character_sync.apply_async(
            kwargs={"pk": character_pk}, priority=DEFAULT_TASK_PRIORITY
        )


@shared_task(base=QueueOnce, bind=True)
@rate_limit_retry_task
def run_character_sync(_self, pk: int):
    """updates in-game contacts for given character

    Args:
    - sync_char_pk: primary key of sync character to run sync for
    """
    sc = SyncedCharacter.objects.get(pk=pk)
    sc.run_sync()


@shared_task
def character_delete_all_contacts(pk: int):
    """Delete contacts of this character."""
    sc = SyncedCharacter.objects.get(pk=pk)
    sc.delete_all_contacts()


@shared_task(base=QueueOnce, bind=True, once={"timeout": ONCE_TIMEOUT})
@rate_limit_retry_task
def sync_all_wars(_self):
    """Sync all wars from ESI."""
    EveWar.objects.sync_known_wars()
    if EveWar.objects.needs_update().exists():
        sync_stale_war.apply_async(priority=DEFAULT_TASK_PRIORITY)


@shared_task(
    base=QueueOnce, bind=True, max_retries=None, once={"timeout": ONCE_TIMEOUT}
)
@rate_limit_retry_task
def sync_stale_war(self: Task):
    """Update the newest stale war from ESI."""
    wars_to_update: QuerySet[EveWar] = EveWar.objects.needs_update()
    war: EveWar = wars_to_update.order_by("-id").first()
    if not war:
        return

    logger.info(
        "%d stale wars need to be updated. Starting to update war ID %d",
        wars_to_update.count(),
        war.id,
    )
    war.update_from_esi()
    logger.info("Updated war with ID %d", war.id)

    delay = SYNC_WAR_DELAY
    logger.debug("Waiting %f seconds for next rate limit slot", delay)
    time.sleep(delay)

    if EveWar.objects.needs_update().exists():
        self.retry(countdown=0.01)



# Generated by Django 5.2.14 on 2026-06-08 12:56

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("standingssync", "0003_allow_empty_evewar_objects"),
    ]

    operations = [
        migrations.AddField(
            model_name="syncmanager",
            name="compress_contacts",
            field=models.BooleanField(
                default=False,
                help_text="Removes contacts which are unnecessary for calculating their effective standing.",
            ),
        ),
        migrations.AddField(
            model_name="syncmanager",
            name="exclude_character_contacts",
            field=models.BooleanField(
                default=False, help_text="Excludes all character contacts"
            ),
        ),
        migrations.AddField(
            model_name="syncmanager",
            name="exclude_negative_standings",
            field=models.BooleanField(
                default=False, help_text="Excludes all contacts with negative standing"
            ),
        ),
        migrations.AddField(
            model_name="syncmanager",
            name="exclude_positive_standings",
            field=models.BooleanField(
                default=False, help_text="Excludes all contacts with positive standing"
            ),
        ),
    ]



"""Alliance Auth app for cloning alliance standings and war targets to alts."""

# pylint: disable = invalid-name
default_app_config = "standingssync.apps.SyncAltContactsConfig"

__version__ = "3.0.0b3"
__title__ = "Standings Sync"



"""Views for standingssync."""

from django.contrib import messages
from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required, permission_required
from django.db.models import Prefetch
from django.shortcuts import get_object_or_404, redirect, render
from esi.decorators import token_required

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveAllianceInfo, EveCharacter
from allianceauth.services.hooks import get_extension_logger

from standingssync import __title__, tasks
from standingssync.app_settings import (
    STANDINGSSYNC_ADD_WAR_TARGETS,
    STANDINGSSYNC_CHAR_MIN_STANDING,
    STANDINGSSYNC_REPLACE_CONTACTS,
    STANDINGSSYNC_WAR_TARGETS_LABEL_NAME,
)
from standingssync.constants import MAX_CHARACTER_CONTACTS
from standingssync.models import EveWar, SyncedCharacter, SyncManager

logger = get_extension_logger(__name__)

MY_DATETIME_FORMAT = "Y-M-d H:i"


def common_context(ctx: dict) -> dict:
    """Return common context used by several views."""
    result = {
        "app_title": __title__,
        "page_title": "PLACEHOLDER",
        "DATEFORMAT": MY_DATETIME_FORMAT,
        "STANDINGSSYNC_ADD_WAR_TARGETS": STANDINGSSYNC_ADD_WAR_TARGETS,
    }
    result.update(ctx)
    return result


@login_required
@permission_required("standingssync.add_syncedcharacter")
def index(request):
    """Render index page."""
    return redirect("standingssync:characters")


@login_required
@permission_required("standingssync.add_syncedcharacter")
def characters(request):
    """Render main page."""
    sync_manager: SyncManager = SyncManager.objects.fetch_for_user(request.user)
    synced_characters = []
    if sync_manager:
        qs = sync_manager.synced_characters_for_user(request.user).select_related(
            "character_ownership", "character_ownership__character"
        )
        for synced_character in qs:
            character = synced_character.character
            organization = str(character.corporation_name)
            if character.alliance_ticker:
                organization += f" [{character.alliance_ticker}]"

            errors = []
            if not synced_character.is_sync_fresh:
                errors.append("Sync is outdated.")
            if (
                STANDINGSSYNC_ADD_WAR_TARGETS
                and synced_character.has_war_targets_label is False
            ):
                errors.append(
                    f"Please create a contact label with the name: "
                    f"{STANDINGSSYNC_WAR_TARGETS_LABEL_NAME}"
                )
            synced_characters.append(
                {
                    "id": character.character_id,
                    "name": character.character_name,
                    "portrait_url": character.portrait_url,
                    "organization": organization,
                    "errors": errors,
                    "pk": synced_character.pk,
                }
            )

        alliance = sync_manager.alliance
        alliance_contacts_count = (
            sync_manager.contacts.filter(is_war_target=False).count()  # type: ignore
            if STANDINGSSYNC_REPLACE_CONTACTS
            else None
        )
        alliance_war_targets_count = (
            sync_manager.contacts.filter(is_war_target=True).count()  # type: ignore
            if STANDINGSSYNC_ADD_WAR_TARGETS
            else None
        )

    else:
        alliance = None
        alliance_contacts_count = None
        alliance_war_targets_count = None

    total_contacts = (alliance_contacts_count or 0) + (alliance_war_targets_count or 0)

    context = {
        "page_title": "My Characters",
        "synced_characters": synced_characters,
        "alliance": alliance,
        "has_synced_chars": len(synced_characters) > 0,
        "alliance_contacts_count": alliance_contacts_count,
        "alliance_war_targets_count": alliance_war_targets_count,
        "war_targets_label_name": STANDINGSSYNC_WAR_TARGETS_LABEL_NAME,
        "overflow": total_contacts > MAX_CHARACTER_CONTACTS,
    }

    return render(request, "standingssync/characters.html", common_context(context))


@login_required
@permission_required("standingssync.add_syncmanager")
@token_required(SyncManager.get_esi_scopes())  # type: ignore
def add_alliance_manager(request, token):
    """Add or update sync manager for an alliance."""
    token_char = get_object_or_404(EveCharacter, character_id=token.character_id)
    if not token_char.alliance_id:
        messages.warning(
            request,
            f"Can not add {token_char}, because it is not a member of any alliance.",
        )
    else:
        character_ownership = get_object_or_404(
            CharacterOwnership, user=request.user, character=token_char
        )
        try:
            alliance = EveAllianceInfo.objects.get(alliance_id=token_char.alliance_id)
        except EveAllianceInfo.DoesNotExist:
            alliance = EveAllianceInfo.objects.create_alliance(token_char.alliance_id)
            alliance.save()
        sync_manager, _ = SyncManager.objects.update_or_create(
            alliance=alliance, defaults={"character_ownership": character_ownership}
        )
        tasks.run_manager_sync.delay(sync_manager.pk)
        if STANDINGSSYNC_ADD_WAR_TARGETS:
            tasks.sync_all_wars.delay()
        messages.success(
            request,
            f"{sync_manager.character.character_name} "
            f"set as alliance character for {alliance.alliance_name}. "
            "Started syncing of alliance contacts. ",
        )
    return redirect("standingssync:index")


@login_required
@permission_required("standingssync.add_syncedcharacter")
@token_required(scopes=SyncedCharacter.get_esi_scopes())  # type: ignore
def add_character(request, token):
    """add character to receive alliance contacts"""
    alliance = get_object_or_404(
        EveAllianceInfo, alliance_id=request.user.profile.main_character.alliance_id
    )
    sync_manager = get_object_or_404(SyncManager, alliance=alliance)
    token_char = get_object_or_404(EveCharacter, character_id=token.character_id)
    if token_char.alliance_id == sync_manager.character.alliance_id:
        messages.warning(
            request,
            "Adding alliance members does not make much sense, "
            "since they already have access to alliance contacts.",
        )

    else:
        character_ownership = get_object_or_404(
            CharacterOwnership, user=request.user, character=token_char
        )
        eff_standing = sync_manager.effective_standing_with_character(
            character_ownership.character
        )
        if eff_standing < STANDINGSSYNC_CHAR_MIN_STANDING:
            messages.warning(
                request,
                "Can not activate sync for your "
                f"character {token_char.character_name}, "
                "because it does not have blue standing "
                "with the alliance. "
                f"The standing value is: {eff_standing:.1f}. "
                "Please first obtain blue "
                "standing for your character and then try again.",
            )
        else:
            sync_character, _ = SyncedCharacter.objects.update_or_create(
                character_ownership=character_ownership,
                defaults={"manager": sync_manager},
            )
            tasks.run_character_sync.delay(sync_character.pk)
            messages.success(
                request, f"Sync activated for {token_char.character_name}!"
            )
    return redirect("standingssync:characters")


@login_required
@permission_required("standingssync.add_syncedcharacter")
def remove_character(request, alt_pk):
    """remove character from receiving alliance contacts"""
    alt = get_object_or_404(SyncedCharacter, pk=alt_pk)
    alt_name = alt.character_ownership.character.character_name
    alt.delete()
    messages.success(request, f"Sync deactivated for {alt_name}")
    return redirect("standingssync:characters")


@login_required
@permission_required("standingssync.add_syncedcharacter")
def wars(request):
    """Render wars page."""
    sync_manager = SyncManager.objects.fetch_for_user(request.user)
    all_wars = []
    if sync_manager:
        query = (
            EveWar.objects.current_wars()
            .alliance_wars(alliance=sync_manager.alliance)
            .prefetch_related(Prefetch("allies", to_attr="allies_sorted"))
            .select_related("aggressor", "defender")
            .annotate_state()
            .annotate_is_active()
            .order_by("-started")
        )
        for war in query:
            allies = sorted(list(war.allies_sorted), key=lambda o: o.name)
            all_wars.append(
                {
                    "declared": war.declared,
                    "started": war.started,
                    "finished": war.finished,
                    "aggressor": war.aggressor,
                    "defender": war.defender,
                    "allies": allies,
                    "state": war.state,
                    "is_active": war.is_active,
                }
            )

    context = {
        "page_title": "Current Wars",
        "alliance": "Not configured",
        "wars": all_wars,
        "war_count": "?",
        "active_wars_count": "?",
        "State": EveWar.State,
    }

    if sync_manager:
        context.update(
            {
                "alliance": sync_manager.alliance,
                "war_count": len(all_wars),
                "active_wars_count": sum(1 for obj in all_wars if obj["is_active"]),
            }
        )

    return render(request, "standingssync/wars.html", common_context(context))


@login_required
@staff_member_required
def admin_update_wars(request):
    """Start updating eve wars."""
    wars_count = EveWar.objects.count()
    tasks.sync_all_wars.delay()
    messages.info(
        request,
        (
            f"Started updating approx. {wars_count:,} wars from ESI in the background. "
            "This can take a minute."
        ),
    )
    return redirect("admin:standingssync_evewar_changelist")



"""Template tags for standingssync."""

from django import template
from eveuniverse.core import zkillboard
from eveuniverse.models import EveEntity

register = template.Library()


@register.inclusion_tag("standingssync/partial/war_participant.html")
def war_participant(obj: EveEntity, icon_size=None) -> dict:
    """Render war participant with icon, name and link to ZKB."""
    context = {"has_data": False, "icon_size": icon_size or 20}
    try:
        category = obj.category
    except AttributeError:
        return context

    if not category:
        return context

    if obj.is_alliance:
        profile_url = zkillboard.alliance_url(obj.id)
    elif obj.is_corporation:
        profile_url = zkillboard.corporation_url(obj.id)
    else:
        profile_url = ""
    context.update(
        {
            "has_data": True,
            "name": obj.name,
            "profile_url": profile_url,
            "icon_url": obj.icon_url(32),
        }
    )
    return context
